
import logging

logger = logging.getLogger(__name__)
logger.warning("y.utils.dank_mids.dank_w3 has been deprecated and will be removed soon. Please import and use `dank_web3` from `dank_mids` directly.")
from dank_mids import dank_web3 as dank_w3



import click

from y import ENVIRONMENT_VARIABLES as ENVS
from y._db import SQLITE_PATH, delete_sqlite
from y._db.config import connection_settings

class CacheNotPopulatedError(Exception):
    pass


class yDBError(Exception):
    provider = ENVS.DB_PROVIDER
    location = SQLITE_PATH if provider == "sqlite" else connection_settings

class NewDatabaseSchemaError(yDBError):
    msg = f"\n\nA more performant ydb schema has been pooshed to prod, you must delete your ypricemagic database at {yDBError.location}"
    def __init__(self) -> None:
        if self.provider == "sqlite" and self._user_approves_sqlite_deletion():
            delete_sqlite()
        super().__init__(self.msg)

    def _user_approves_sqlite_deletion(self) -> bool:
        return click.prompt(
            self.msg + "\n\nWould you like to delete your db now?", 
            default="no", 
            type=click.Choice(["yes", "no"], case_sensitive=False), 
            confirmation_prompt=True, 
            show_default=False,
        ) == "yes"

class EEEError(ValueError):
    """raised when trying to do stuff with the eee address in the token db"""



import os

from pony.orm import Database, BindingError, DatabaseError, TransactionError

from y._db.config import SQLITE_PATH


def bind_db(db: Database, **connection_settings) -> None:
    try:
        db.bind(**connection_settings)
    except BindingError as e:
        if not str(e).startswith('Database object was already bound to'):
            raise

def generate_mapping(db: Database) -> None:
    try:
        db.generate_mapping(create_tables=True)
    except BindingError as e:
        if str(e) != "Mapping was already generated":
            raise
    except DatabaseError as e:
        if "no such column: " in str(e) or ("column" in str(e) and " does not exist" in str(e)):
            from y._db.exceptions import NewDatabaseSchemaError
            raise NewDatabaseSchemaError from e
    except TransactionError as e:
        if str(e) != "@db_session-decorated create_tables() function with `ddl` option cannot be called inside of another db_session":
            raise
    
def delete_sqlite() -> None:
    os.remove(SQLITE_PATH)




import logging
from itertools import chain
from typing import AsyncIterator, List, Optional

import a_sync
import dank_mids
from a_sync.executor import _AsyncExecutorMixin
from msgspec import json
from pony.orm import commit, select

from y._db.common import DiskCache, Filter, _clean_addresses, default_filter_threads
from y._db.decorators import a_sync_write_db_session
from y._db.entities import Chain, Trace, TraceCacheInfo, insert
from y._db.utils._ep import _get_get_block
from y.utils.middleware import BATCH_SIZE

logger = logging.getLogger(__name__)

@a_sync_write_db_session
def insert_trace(trace: dict) -> None:
    get_block = _get_get_block()
    kwargs = {
        "block": get_block(trace['blockNumber'], sync=True),
        "hash": trace['transactionHash'],
        "raw": json.encode(trace),
    }
    for dct in [trace, *trace.values()]:
        if "from" in dct:
            kwargs['from_address'] = dct["from"]
        if "to" in dct:
            kwargs['to_address'] = dct["to"]
    insert(type=Trace, **kwargs)

class TraceCache(DiskCache[dict, TraceCacheInfo]):
    __slots__ = "from_addresses", "to_addresses"
    def __init__(self, from_addresses: List[str], to_addresses: List[str]):
        self.from_addresses = _clean_addresses(from_addresses)
        self.to_addresses = _clean_addresses(to_addresses)
    
    def load_metadata(self, chain: Chain, from_address: Optional[str], to_address: Optional[str]) -> Optional[TraceCacheInfo]:
        return TraceCacheInfo.get(chain=chain, from_address=str(from_address), to_address=str(to_address))
    
    def _is_cached_thru(self, from_block: int) -> int:
        from y._db.utils import utils as db
        chain = db.get_chain(sync=True)
        infos = [
            self.load_metadata(chain, from_address, to_address)
            or self.load_metadata(chain, from_address, None)
            or self.load_metadata(chain, None, to_address)
            for from_address in self.from_addresses
            for to_address in self.to_addresses
        ]
        if all(info and info.cached_from <= from_block for info in infos):
            return max(info.cached_thru for info in infos)
        return 0

    def _select(self, from_block: int, to_block: int) -> List[dict]:
        from y._db.utils import utils as db
        return [
            json.decode(trace) for trace in select(
                trace.raw for trace in Trace 
                if trace.block.chain == db.get_chain(sync=True)
                and (not self.from_addresses or trace.address in self.from_addresses)
                and (not self.to_addresses or trace.address in self.to_addresses)
                and trace.block.number >= from_block
                and trace.block.number <= to_block
            )
        ]
    
    def _set_metadata(self, from_block: int, done_thru: int) -> None:
        from y._db.utils import utils as db
        chain = db.get_chain(sync=True)
        should_commit = False
        if self.to_addresses and self.from_addresses:
            for from_address in self.from_addresses:
                for to_address in self.to_addresses:
                    if info := TraceCacheInfo.get(
                            chain=chain, 
                            from_address=json.encode([from_address]),
                            to_address=json.encode([to_address]),
                        ):
                            if from_block < info.cached_from:
                                info.cached_from = from_block
                                should_commit = True
                            if done_thru > info.cached_thru:
                                info.cached_thru = done_thru
                                should_commit = True
                    else:
                        TraceCacheInfo(
                            chain=chain, 
                            from_address=json.encode([from_address]),
                            to_address=json.encode([to_address]),
                        )
                        should_commit = True
        elif self.from_addresses:
            for from_address in self.from_addresses:
                if info := TraceCacheInfo.get(
                    chain=chain, 
                    from_address=json.encode([from_address]),
                    to_address=json.encode([]),
                ):
                    if from_block < info.cached_from:
                        info.cached_from = from_block
                        should_commit = True
                    if done_thru > info.cached_thru:
                        info.cached_thru = done_thru
                        should_commit = True
                else:
                    TraceCacheInfo(
                        chain=chain, 
                        from_address=json.encode([from_address]),
                        to_address=json.encode([]),
                    )
                    should_commit = True
        elif self.to_addresses:
            for to_address in self.to_addresses:
                if info := TraceCacheInfo.get(
                    chain=chain, 
                    from_address=json.encode([]),
                    to_address=json.encode([to_address]),
                ):
                    if from_block < info.cached_from:
                        info.cached_from = from_block
                        should_commit = True
                    if done_thru > info.cached_thru:
                        info.cached_thru = done_thru
                        should_commit = True
                else:
                    TraceCacheInfo(
                        chain=chain, 
                        from_address=json.encode([]),
                        to_address=json.encode([to_address]),
                    )
                    should_commit = True
        elif info := TraceCacheInfo.get(
            chain=chain, 
            from_address=json.encode([]),
            to_address=json.encode([])
        ):
            if from_block < info.cached_from:
                info.cached_from = from_block
                should_commit = True
            if done_thru > info.cached_thru:
                info.cached_thru = done_thru
                should_commit = True
        else:
            TraceCacheInfo(
                chain=chain, 
                from_address=json.encode([]),
                to_address=json.encode([])
            )
            should_commit = True

        if should_commit:
            commit()
            logger.debug('cached %s %s thru %s', self.from_addresses, self.to_addresses, done_thru)
        


class TraceFilter(Filter[dict, TraceCache]):
    insert_to_db = insert_trace
    __slots__ = "from_addresses", "to_addresses"
    def __init__(
        self, 
        from_addresses: List[str], 
        to_addresses: List[str], 
        from_block: int,
        *,
        chunk_size: int = BATCH_SIZE,
        chunks_per_batch: Optional[int] = None,
        semaphore: Optional[dank_mids.BlockSemaphore] = None,
        executor: _AsyncExecutorMixin = default_filter_threads,
        is_reusable: bool = True,
        verbose: bool = False,
    ):
        self.from_addresses = from_addresses
        self.to_addresses = to_addresses
        super().__init__(from_block, chunk_size=chunk_size, chunks_per_batch=chunks_per_batch, semaphore=semaphore, executor=executor, is_reusable=is_reusable, verbose=verbose)

    @property
    def cache(self) -> TraceCache:
        if self._cache is None:
            self._cache = TraceCache(self.from_addresses, self.to_addresses)
        return self._cache

    def traces(self, to_block: Optional[int]) -> AsyncIterator[dict]:
        return self._objects_thru(block=to_block)

    async def _fetch_range(self, from_block: int, to_block: int) -> List[dict]:
        try:
            return await dank_mids.web3.provider.make_request("TraceFilter", {})
        except NotImplementedError:
            results = {block: traces async for block, traces in a_sync.map(self._trace_block, range(from_block, to_block)).map()}
            return list(chain(*[results[i] for i in range(from_block, to_block)]))
        
    async def _trace_block(self, block: int) -> List[dict]:
        return [
            trace for trace in await dank_mids.web3.provider.make_request("TraceBlock", block)
            if (not self.from_addresses or any("from" in x and x["from"] in self.from_addresses for x in [trace, trace.values()]))
            and (not self.to_addresses or any("to" in x and x["to"] in self.to_addresses for x in [trace, trace.values()]))
        ]



import logging
from typing import Optional

import a_sync
from brownie import chain

from y import convert
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.networks import Network
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)

POOL = '0x83E5f18Da720119fF363cF63417628eB0e9fd523'

def is_froyo(token: AnyAddressType) -> bool:
    address = convert.to_address(token)
    return chain.id == Network.Fantom and address == '0x4f85Bbf3B0265DCEd4Ec72ebD0358ccCf190F1B3'

@a_sync.a_sync(default='sync')
async def get_price(token: AnyAddressType, block: Optional[Block] = None) -> Optional[UsdPrice]:
    if not token == POOL:
        return None
    virtual_price = await raw_call(POOL, "get_virtual_price()", block=block, output='int', sync=False)
    return UsdPrice(virtual_price / 1e18)




import y
from tests.fixtures import async_test, optimism_only


@async_test
@optimism_only
async def test_wtbt():
    assert await y.get_price("0xdb4eA87fF83eB1c80b8976FC47731Da6a31D35e5", 106500000, skip_cache=True, sync=False) == 1.0009538615847542

@async_test
@optimism_only
async def test_mseth():
    assert await y.get_price("0x1610e3c85dd44Af31eD7f33a63642012Dca0C5A5", 106500000, skip_cache=True, sync=False) == 1909.2991854198503



from typing import Iterable, List, Tuple

import numpy as np
import pytest
from brownie import Contract as BrownieContract
from brownie import chain
from brownie.convert.datatypes import EthAddress
from y import convert
from y.classes.common import ERC20
from y.contracts import Contract, contract_creation_block
from y.datatypes import Address, Block
from y.prices.dex.uniswap.v1 import UniswapV1

from y.networks import Network

mainnet_only = pytest.mark.skipif(
    chain.id != Network.Mainnet, reason="This test is only applicable on mainnet"
)
"""
A pytest marker to skip tests that are only applicable on the Ethereum mainnet.
"""

optimism_only = pytest.mark.skipif(
    chain.id != Network.Optimism, reason="This test is only applicable on optimism"
)
"""
A pytest marker to skip tests that are only applicable on the Optimism network.
"""

async_test = pytest.mark.asyncio_cooperative
"""
A pytest marker for concurrent asynchronous tests using asyncio_cooperative.
"""

def blocks_for_contract(address: Address, count: int = 5) -> List[Block]:
    """
    Generate a list of block numbers for testing a contract.

    Args:
        address: The address of the contract.
        count: The number of blocks to generate (default is 5).

    Returns:
        A list of block numbers evenly spaced between the contract's creation and the current block height.
    """
    address = convert.to_address(address)
    return [int(block) for block in np.linspace(contract_creation_block(address) + 10000, chain.height, count)]

def mutate_address(address: Address) -> Tuple[str,str,str,EthAddress]:
    """
    Returns the same address in various forms for testing.

    Args:
        address: The original address.

    Returns:
        A tuple containing the address in four different forms:
        1. Lowercase string
        2. Checksum address string
        3. Address object
        4. EthAddress object
    """
    return (
        address.lower(),
        address[:2] + address[2:].upper(),
        convert.to_address(address),
        EthAddress(address)
    )

def mutate_addresses(addresses: Iterable[Address]):
    return [mutation for address in addresses for mutation in mutate_address(address)]

def mutate_contract(contract_address: Address) -> Tuple[str,str,str,EthAddress,BrownieContract]:
    """
    Returns the same contract address in various forms for testing.
    """
    mutations = list(mutate_address(contract_address))
    mutations.append(Contract(contract_address))
    return tuple(mutations)

def mutate_contracts(addresses: Iterable[Address]):
    return [mutation for address in addresses for mutation in mutate_contract(address)]

def mutate_token(token: Address) -> Tuple[str,str,str,EthAddress,BrownieContract,int,ERC20]:
    """
    Returns the same token address in various forms for testing.
    """
    mutations = list(mutate_contract(token))
    mutations.append(ERC20(token))
    mutations.append(int(token,16))
    return tuple(mutations)

def mutate_tokens(addresses: Iterable[Address]):
    return [mutation for address in addresses for mutation in mutate_token(address)]

@pytest.fixture
def async_uni_v1():
    yield UniswapV1(asynchronous=True)


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

from sphinx.util import logging


project = 'ypricemagic'
copyright = '2024, BobTheBuidler'
author = 'BobTheBuidler'

logger = logging.getLogger(__name__)

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
    'sphinx.ext.intersphinx',
    'a_sync.sphinx.ext',
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# Looks for objects in documentation of external libraries
intersphinx_mapping = {
    'a_sync': ('https://bobthebuidler.github.io/ez-a-sync', None),
    'brownie': ('https://eth-brownie.readthedocs.io/en/stable/', None),
    'checksum_dict': ('https://bobthebuidler.github.io/checksum_dict', None),
    'dank_mids': ('https://bobthebuidler.github.io/dank_mids', None),
    'eth_typing': ('https://eth-typing.readthedocs.io/en/stable/', None),
    'hexbytes': ('https://hexbytes.readthedocs.io/en/stable/', None),
    'pony': ('https://docs.ponyorm.org/', None),
    'python': ('https://docs.python.org/3', None),
    'typing_extensions': ('https://typing-extensions.readthedocs.io/en/latest/', None),
    'web3': ('https://web3py.readthedocs.io/en/stable/', None),
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']

autodoc_default_options = {
    'undoc-members': True,
    'special-members': True,
    'inherited-members': True,
    'member-order': 'groupwise',
    # hide private methods that aren't relevant to us here
    'exclude-members': ','.join([
        '__new__',
        '__dict__',
        '__slots__',
        '__module__',
        '__hash__',
        '__repr__',
        '__setattr__',
        '__delattr__',
        '__init_from_abi__',
        '__method_names__',
        '__get_method_object__',
        '__annotations__',
        '__weakref__',
        '__reduce__',
        '__cause__',
        '__context__',
        '__suppress_context__',
        '__setstate__',
        '__traceback__',
        '_abc_impl',
        '__abstractmethods__',
        '__a_sync_default_mode__',
        '__a_sync_instance_will_be_sync__',
        '__a_sync_should_await__',
        '__a_sync_should_await_from_kwargs__',
        '__a_sync_flag_name__',
        '__a_sync_flag_value__',
        '__a_sync_instance_should_await__',
        '__a_sync_modifiers__',
        '__class_getitem__',
        '__init_subclass__',
        '__subclass_hook__',
    ]),
}
autodoc_typehints = "description"
# Don't show class signature with the class' name.
autodoc_class_signature = "separated"

automodule_generate_module_stub = True

sys.path.insert(0, os.path.abspath('./y'))


def skip_specific_members(app, what, name, obj, skip, options):
    """
    Function to exclude specific members for a particular module.
    """
    
    logger.info(f"module: {getattr(obj, '__module__', None)}  name: {name}  obj: {obj}")

    # Skip the __init__ and __call__ members of any NewType objects we defined.
    if type(getattr(obj, "__self__", None)).__qualname__ == "typing.NewType" and name in ["__init__", "__call__"]:
        return True
    
    # Skip the __init__, __str__, __getattribute__, args, and with_traceback members of all Exceptions
    if issubclass(getattr(obj, '__objclass__', type), BaseException) and name in ["__init__", "__str__", "__getattribute__", "args", "with_traceback"]:
        return True
    
    return skip

def setup(app):
    app.connect('autodoc-skip-member', skip_specific_members)


import logging
import os
from typing import Any, Callable

import eth_retry
from brownie import chain, web3
from requests import Session
from requests.adapters import HTTPAdapter
from web3 import HTTPProvider, Web3
from web3.middleware import filter
from web3.middleware.geth_poa import geth_poa_middleware

from y import ENVIRONMENT_VARIABLES as ENVS
from y.networks import Network
from y.utils.cache import memory

logger = logging.getLogger(__name__)

provider_specific_batch_sizes = {
    "moralis":  2_000,
    "pokt":     2_000,
    "tenderly": 2_000,
    "ankr":     2_000,
}

chain_specific_max_batch_sizes = {
    Network.Mainnet:    10_000,     # 1.58 days
    Network.Gnosis:     20_000,     # 1.15 days
    Network.Fantom:     100_000,    # 1.03 days
    Network.Arbitrum:   20_000,     # 0.34 days
    Network.Optimism:   800_000,    # 10.02 days
}

fallback_batch_size = 10_000

def _get_batch_size() -> int:
    """
    Determine the appropriate batch size for the current network and provider.

    Returns:
        The determined batch size for log fetching.
    """
    if batch_size := ENVS.GETLOGS_BATCH_SIZE:
        return batch_size
    for provider, size in provider_specific_batch_sizes.items():
        if provider in web3.provider.endpoint_uri:
            return size
    return chain_specific_max_batch_sizes.get(chain.id, fallback_batch_size)

BATCH_SIZE = _get_batch_size()


def should_cache(method: str, params: Any) -> bool:
    """
    Determine if a method call should be cached.

    Args:
        method: The name of the method being called.
        params: The parameters of the method call.

    Returns:
        True if the method call should be cached, False otherwise.
    """
    return method == "eth_getCode" and params[1] == "latest"


def getcode_cache_middleware(make_request: Callable, web3: Web3) -> Callable:
    """
    Middleware for caching eth_getCode calls.

    Args:
        make_request: The original request function.
        web3: The Web3 instance.

    Returns:
        A middleware function that caches eth_getCode calls.
    """
    @eth_retry.auto_retry
    def middleware(method: str, params: Any) -> Any:
        logger.debug("%s %s", method, params)
        if should_cache(method, params):
            return memory.cache(make_request)(method, params)
        return make_request(method, params)
    return middleware


def setup_getcode_cache_middleware() -> None:
    """
    Set up the eth_getCode cache middleware for the current Web3 provider.

    This function modifies the Web3 provider to use a custom session with increased
    connection pool size and timeout, and adds the getcode cache middleware.

    On Optimism, it also adds the POA middleware.
    """
    # patch web3 provider with more connections and higher timeout
    if web3.provider:
        try:
            assert web3.provider.endpoint_uri.startswith("http"), "only http and https providers are supported"
            adapter = HTTPAdapter(pool_connections=100, pool_maxsize=100)
            session = Session()
            session.mount("http://", adapter)
            session.mount("https://", adapter)
            web3.provider = HTTPProvider(web3.provider.endpoint_uri, {"timeout": 600}, session)
        except AttributeError as e:
            if "'IPCProvider' object has no attribute 'endpoint_uri'" not in str(e):
                raise

    web3.middleware_onion.add(getcode_cache_middleware)
    
    if chain.id == Network.Optimism:
        try:
            web3.middleware_onion.inject(geth_poa_middleware, layer=0)
        except ValueError as e:
            if str(e) != "You can't add the same un-named instance twice":
                raise




import asyncio
import logging
from decimal import Decimal
from typing import List, Optional

import a_sync
from multicall import Call

from y import ENVIRONMENT_VARIABLES as ENVS
from y.classes.common import ERC20, WeiBalance
from y.contracts import Contract, has_methods
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.utils.cache import optional_async_diskcache

logger = logging.getLogger(__name__)


@a_sync.a_sync(default='sync', cache_type='memory', ram_cache_ttl=5*60)
@optional_async_diskcache
async def is_token_set(token: AnyAddressType) -> bool:
    return any(await asyncio.gather(
        has_methods(token, ("getComponents()(address[])", "naturalUnit()(uint)"), sync=False),
        has_methods(token, ("getComponents()(address[])", "getModules()(address[])", "getPositions()(address[])"), sync=False),
    ))

@a_sync.a_sync(default='sync')
async def get_price(token: AnyAddressType, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> UsdPrice:
    return await TokenSet(token, asynchronous=True).get_price(block=block, skip_cache=skip_cache)


class TokenSet(ERC20):
    def __init__(self, address: AnyAddressType, asynchronous: bool = False):
        super().__init__(address, asynchronous=asynchronous)
        self.get_total_component_real_units = Call(self.address, "getTotalComponentRealUnits(address)(int256)")

    @a_sync.a_sync(ram_cache_maxsize=100)
    async def components(self, block: Optional[Block] = None) -> List[ERC20]:
        contract = await Contract.coroutine(self.address)
        components = await contract.getComponents.coroutine(block_identifier = block)
        return [ERC20(component, asynchronous=self.asynchronous) for component in components]
    
    async def balances(self, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> List[WeiBalance]:
        contract, components = await asyncio.gather(
            Contract.coroutine(self.address),
            self.components(block=block, sync=False),
        )
        if hasattr(contract, 'getUnits'):
            balances = await contract.getUnits.coroutine(block_identifier = block)
            logger.debug("getUnits: %s", balances)
        elif hasattr(contract, 'getTotalComponentRealUnits'):
            _components = ((component.address, ) for component in components)
            balances = await a_sync.map(self.get_total_component_real_units.coroutine, _components, block_id=block).values(pop=True)
            logger.debug("getTotalComponentRealUnits: %s", balances)
        balances = [WeiBalance(balance, component, block=block, skip_cache=skip_cache) for component, balance in zip(components, balances)]
        return balances
    
    async def get_price(self, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> UsdPrice:
        total_supply = Decimal(await self.total_supply_readable(block=block, sync=False))
        if total_supply == 0:
            logger.debug("total supply is 0, forcing price to $0")
            return UsdPrice(0)
        contract = await Contract.coroutine(self.address)
        if hasattr(contract, "getUnits"):
            balances: List[WeiBalance] = await self.balances(block=block, skip_cache=skip_cache, sync=False)
            values = await WeiBalance.value_usd.map(balances).values(pop=True)
            logger.debug("balances: %s  values: %s", balances, values)
            tvl = sum(values)
            price = UsdPrice(tvl / total_supply)
            logger.debug("total supply: %s  tvl: %s  price: %s", total_supply, tvl, price)
            return price
        elif hasattr(contract, "getTotalComponentRealUnits"):
            balances_per_token: List[WeiBalance] = await self.balances(block=block, sync=False)
            price = UsdPrice(await WeiBalance.value_usd.sum(balances_per_token, sync=False))
            logger.debug("balances per token: %s  price: %s", balances_per_token, price)
            return price
        raise NotImplementedError



"""
Utility functions for retrieving Ethereum client information.
"""

import logging

from async_lru import alru_cache
from brownie import web3
from cachetools.func import lru_cache
from dank_mids import dank_web3
from web3._utils.rpc_abi import RPC

from y.utils.logging import yLazyLogger

logger = logging.getLogger(__name__)

@lru_cache(1)
@yLazyLogger(logger)
def get_ethereum_client() -> str:
    """
    Get the Ethereum client type for the current connection.

    Returns:
        A string representing the Ethereum client type.

    Example:
        >>> from y.utils.client import get_ethereum_client
        >>> client = get_ethereum_client()
        >>> print(client)
        'geth'
    """
    return _get_ethereum_client(web3.clientVersion)

@alru_cache(maxsize=1)
async def get_ethereum_client_async() -> str:
    """
    Asynchronously get the Ethereum client type for the current connection.

    Returns:
        A string representing the Ethereum client type.

    Example:
        >>> from y.utils.client import get_ethereum_client_async
        >>> client = await get_ethereum_client_async()
        >>> print(client)
        'erigon'
    """
    return _get_ethereum_client(await dank_web3.manager.coro_request(RPC.web3_clientVersion, []))

@yLazyLogger(logger)
def _get_ethereum_client(client: str) -> str:
    """
    Determine the Ethereum client type based on the client version string.

    Args:
        client: A string containing the client version information.

    Returns:
        A string representing the Ethereum client type.

    Example:
        >>> from y.utils.client import _get_ethereum_client
        >>> client_type = _get_ethereum_client("Geth/v1.10.23-stable/linux-amd64/go1.18.7")
        >>> print(client_type)
        'geth'
    """
    logger.debug("client: %s", client)
    return next((identifier for prefix, identifier in _clients.items() if client.startswith(prefix)), client)

_clients = {
    'geth': 'geth',
    'erigon': 'erigon',
    'TurboGeth': 'tg',
}



import logging
from datetime import datetime
from decimal import Decimal
from typing import Any
from typing import Optional as typing_Optional

import a_sync
from pony.orm import (Database, InterfaceError, Optional, PrimaryKey, 
                      Required, Set, TransactionIntegrityError, commit,
                      composite_index, db_session)

from y._db.decorators import retry_locked, ydb_write_threads

db = Database()


logger = logging.getLogger(__name__)

class _AsyncEntityMixin:
    pass
    '''
    @classmethod
    @a_sync('async')
    def get(cls: E, *args, **kwargs) -> E:
        return super(db.Entity).get(*args, **kwargs)
    @classmethod
    @a_sync('async')
    def select(cls: E, *args, **kwargs) -> E:
        return super(db.Entity).select(*args, **kwargs)'''
    
class Chain(db.Entity, _AsyncEntityMixin):
    id = PrimaryKey(int)

    blocks = Set("Block")
    addresses = Set("Address")
    log_cached = Set("LogCacheInfo")
    trace_caches = Set("TraceCacheInfo")

class Block(db.Entity, _AsyncEntityMixin):
    chain = Required(Chain, reverse="blocks")
    number = Required(int)
    PrimaryKey(chain, number)
    hash = Optional(int, lazy=True)
    timestamp = Optional(datetime, lazy=True)
    
    composite_index(chain, hash)

    prices = Set("Price", reverse="block", cascade_delete=False)
    contracts_deployed = Set("Contract", reverse="deploy_block")
    logs = Set("Log", reverse="block", cascade_delete=False)
    traces = Set("Trace", reverse="block", cascade_delete=False)

class Address(db.Entity, _AsyncEntityMixin):
    chain = Required(Chain, lazy=True, reverse="addresses")
    address = Required(str, lazy=True)
    PrimaryKey(chain, address)
    notes = Optional(str, lazy=True)
    
    contracts_deployed = Set("Contract", reverse="deployer")

class Contract(Address):
    deployer = Optional(Address, reverse='contracts_deployed', lazy=True, cascade_delete=False)
    deploy_block = Optional(Block, reverse='contracts_deployed', lazy=True, cascade_delete=False)

class Token(Contract):
    symbol = Optional(str, lazy=True)
    name = Optional(str, lazy=True)
    decimals = Optional(int, lazy=True)
    bucket = Optional(str, index=True, lazy=True)

    prices = Set("Price", reverse="token")

class Price(db.Entity):
    block = Required(Block, index=True, lazy=True)
    token = Required(Token, index=True, lazy=True)
    PrimaryKey(block, token)
    price = Required(Decimal, 38, 18)
    
class TraceCacheInfo(db.Entity):
    chain = Required(Chain, index=True)
    to_addresses = Required(bytes, index=True)
    from_addresses = Required(bytes, index=True)
    PrimaryKey(chain, to_addresses, from_addresses)
    cached_from = Required(int)
    cached_thru = Required(int)

class LogCacheInfo(db.Entity):
    chain = Required(Chain, index=True)
    address = Required(str, index=True)
    topics = Required(bytes)
    PrimaryKey(chain, address, topics)
    cached_from = Required(int)
    cached_thru = Required(int)

class LogTopic(db.Entity):
    "Just makes the :ref:`Log` db smaller."
    dbid = PrimaryKey(int, auto=True, size=64)
    topic = Required(str, unique=True, lazy=True)

    logs_as_topic0 = Set("Log", reverse="topic0")
    logs_as_topic1 = Set("Log", reverse="topic1")
    logs_as_topic2 = Set("Log", reverse="topic2")
    logs_as_topic3 = Set("Log", reverse="topic3")

class Hashes(db.Entity):
    "Just makes :ref:`Log` pk and indexes smaller."
    dbid = PrimaryKey(int, auto=True, size=64)
    hash = Required(str, unique=True)

    logs_for_tx = Set("Log", reverse="tx")
    logs_for_addr = Set("Log", reverse="address")

class Log(db.Entity):
    block = Required(Block, index=True, lazy=True)
    tx = Required(Hashes, lazy=True)
    log_index = Required(int, size=16, lazy=True)
    PrimaryKey(block, tx, log_index)

    address = Required(Hashes, index=True, lazy=True)
    
    topic0 = Required(LogTopic, index=True, lazy=True)
    topic1 = Optional(LogTopic, index=True, lazy=True)
    topic2 = Optional(LogTopic, index=True, lazy=True)
    topic3 = Optional(LogTopic, index=True, lazy=True)
    composite_index(address, topic0)
    composite_index(topic0, topic1)
    composite_index(topic0, topic2)
    composite_index(topic0, topic3)
    composite_index(block, address, topic0)
    composite_index(block, topic0, topic1)
    composite_index(block, topic0, topic2)
    composite_index(block, topic0, topic3)
    composite_index(topic0, topic1, topic2, topic3)
    composite_index(block, topic0, topic1, topic2, topic3)
    raw = Required(bytes, lazy=True)

class Trace(db.Entity):
    id = PrimaryKey(int, auto=True)
    block = Required(Block, index=True, lazy=True)
    hash = Required(str, index=True, lazy=True)
    from_address = Required(str, index=True, lazy=True)
    to_address = Required(str, index=True, lazy=True)
    raw = Required(bytes)

class BlockAtTimestamp(db.Entity):
    chainid = Required(int)
    timestamp = Required(datetime)
    PrimaryKey(chainid, timestamp)
    block = Required(int)


@a_sync.a_sync(executor=ydb_write_threads)
@db_session
@retry_locked
def insert(type: db.Entity, **kwargs: Any) -> typing_Optional[db.Entity]:
    try:
        while True:
            try:
                entity = type(**kwargs)
                commit()
                logger.debug("inserted %s to db", entity)
                return entity
            except InterfaceError as e:
                logger.debug("%s while inserting %s", e, type.__name__)
    except TransactionIntegrityError as e:
        constraint_errs = "UNIQUE constraint failed", "duplicate key value violates unique constraint"
        if any(err in (msg:=str(e)) for err in constraint_errs):
            logger.debug("%s: %s %s", msg, type.__name__, kwargs)
        else:
            logger.debug("%s %s when inserting %s", e.__class__.__name__, str(e), e, type.__name__)
            raise




import abc
import asyncio
from contextlib import suppress
from decimal import Decimal
from functools import cached_property
from logging import getLogger
from typing import (TYPE_CHECKING, Any, Awaitable, Generator, Literal, NoReturn,
                    Optional, Tuple, Union, final)

import a_sync
from a_sync.a_sync import HiddenMethodDescriptor, HiddenMethod
from a_sync.a_sync.method import ASyncBoundMethod
from brownie import Contract, chain, web3
from brownie.convert.datatypes import HexString
from brownie.exceptions import ContractNotFound
from eth_retry import auto_retry
from typing_extensions import Self

from y import ENVIRONMENT_VARIABLES as ENVS
from y import convert
from y._decorators import stuck_coro_debugger
from y.classes.singleton import ChecksumASyncSingletonMeta
from y.constants import EEE_ADDRESS
from y.contracts import (Contract, build_name, contract_creation_block_async,
                         has_method, probe)
from y.datatypes import Address, AnyAddressType, Block, Pool, UsdPrice
from y.erc20 import decimals, totalSupply
from y.exceptions import (ContractNotVerified, MessedUpBrownieContract,
                          NonStandardERC20)
from y.networks import Network
from y.utils import logging, raw_calls

if TYPE_CHECKING:
    from y.utils.events import Events

logger = getLogger(__name__)

def hex_to_string(h: HexString) -> str:
    '''returns a string from a HexString'''
    h = h.hex().rstrip("0")
    if len(h) % 2 != 0:
        h += "0"
    return bytes.fromhex(h).decode("utf-8")

class ContractBase(a_sync.ASyncGenericBase, metaclass=ChecksumASyncSingletonMeta):
    # defaults are stored as class vars to keep instance dicts smaller
    asynchronous: bool = False
    _deploy_block: Optional[int] = None
    __slots__ = "address",
    def __init__(
        self, 
        address: AnyAddressType, 
        asynchronous: bool = False, 
        _deploy_block: Optional[int] = None,
    ) -> None:
        self.address = convert.to_address(address)
        if asynchronous:
            self.asynchronous = asynchronous
        if _deploy_block:
            self._deploy_block = _deploy_block
        super().__init__()
    
    def __str__(self) -> str:
        return f'{self.address}'

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} '{self.address}'"
    
    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, (ContractBase, Contract)):
            return __o.address == self.address
        # Skip checksumming if applicable, its computationally expensive
        # NOTE: We assume a mixed-case address is checksummed. If it isn't, wtf are you doing?
        elif isinstance(__o, str) and __o != __o.lower() and __o != __o.upper():
            return __o == self.address
        try:
            return convert.to_address(__o) == self.address
        except Exception:
            return False
    
    def __hash__(self) -> int:
        return hash(self.address)
    
    @property
    def contract(self) -> Contract:
        return Contract(self.address)
    
    @cached_property
    def _is_cached(self) -> bool:
        try:
            self.contract
            return True
        except (ContractNotVerified):
            return False
        except (ContractNotFound, MessedUpBrownieContract):
            return None

    @a_sync.aka.cached_property
    @stuck_coro_debugger
    async def build_name(self) -> str:
        return await build_name(self.address, sync=False)
    __build_name__: HiddenMethodDescriptor[Self, str]

    @stuck_coro_debugger
    async def deploy_block(self, when_no_history_return_0: bool = False) -> int:
        if self._deploy_block is None:
            self._deploy_block = await contract_creation_block_async(self.address, when_no_history_return_0=when_no_history_return_0)
        return self._deploy_block
    deploy_block: ASyncBoundMethod[Self, Any, int]
    
    async def has_method(self, method: str, return_response: bool = False) -> Union[bool,Any]:
        return await has_method(self.address, method, return_response=return_response, sync=False)


class ERC20(ContractBase):
    def __repr__(self) -> str:
        cls = type(self).__name__
        with suppress(AttributeError):
            if ERC20.symbol.has_cache_value(self):
                symbol = ERC20.symbol.get_cache_value(self)
                return f"<{cls} {symbol} '{self.address}'>"
            elif not asyncio.get_event_loop().is_running() and not self.asynchronous:
                try:
                    return f"<{cls} {self.__symbol__(sync=True)} '{self.address}'>"
                except NonStandardERC20:
                    return f"<{cls} SYMBOL_INVALID '{self.address}'>"
        return f"<{cls} SYMBOL_NOT_LOADED '{self.address}'>"
    
    @a_sync.aka.cached_property
    @stuck_coro_debugger
    async def symbol(self) -> str:
        if self.address == EEE_ADDRESS:
            return {
                Network.Mainnet: "ETH",
                Network.Fantom: "FTM",
                Network.Polygon: "MATIC",
                Network.Arbitrum: "ETH",
                Network.Optimism: "ETH",
                Network.Base: "ETH",
            }.get(chain.id, "ETH")
        import y._db.utils.token as db
        if symbol := await db.get_symbol(self.address):
            return symbol
        symbol = await self._symbol()
        db.set_symbol(self.address, symbol)
        return symbol
    
    @a_sync.aka.property
    @stuck_coro_debugger
    async def name(self) -> str:
        if self.address == EEE_ADDRESS:
            return "Ethereum"
        import y._db.utils.token as db
        name = await db.get_name(self.address)
        if name:
            return name
        name = await self._name()
        db.set_name(self.address, name)
        return name
    
    @a_sync.aka.cached_property
    @stuck_coro_debugger
    async def decimals(self) -> int:
        if self.address == EEE_ADDRESS:
            return 18
        import y._db.utils.token as db
        return await db.get_decimals(self.address)

    @a_sync.a_sync # Override the leading underscore so a_sync lib doesn't bypass this fn
    async def _decimals(self, block: Optional[Block] = None) -> int:
        '''used to fetch decimals at specific block'''
        if self.address == EEE_ADDRESS:
            return 18
        return await decimals(self.address, block=block, sync=False)
    
    @a_sync.aka.cached_property
    @stuck_coro_debugger
    async def scale(self) -> int:
        return 10 ** await self.__decimals__(asynchronous=True)
    
    @a_sync.a_sync # Override the leading underscore so a_sync lib doesn't bypass this fn
    async def _scale(self, block: Optional[Block] = None) -> int:
        return 10 ** await self._decimals(block, sync=False)

    async def total_supply(self, block: Optional[Block] = None) -> int:
        return await totalSupply(self.address, block=block, sync=False)

    async def total_supply_readable(self, block: Optional[Block] = None) -> float:
        total_supply, scale = await asyncio.gather(self.total_supply(block=block, sync=False), self.__scale__)
        return total_supply / scale
    
    async def balance_of(self, address: AnyAddressType, block: Optional[Block] = None) -> int:
        return await raw_calls.balanceOf(self.address, address, block=block, sync=False)
    
    async def balance_of_readable(self, address: AnyAddressType, block: Optional[Block] = None) -> float:
        balance, scale = await asyncio.gather(self.balance_of(address, block=block, asynchronous=True), self.__scale__(asynchronous=True))
        return balance / scale

    async def price(
        self, 
        block: Optional[Block] = None, 
        return_None_on_failure: bool = False, 
        skip_cache: bool = ENVS.SKIP_CACHE,
        ignore_pools: Tuple[Pool, ...] = (),
    ) -> Optional[UsdPrice]:
        from y.prices.magic import get_price
        return await get_price(
            self.address, 
            block=block, 
            fail_to_None=return_None_on_failure,
            skip_cache=skip_cache,
            ignore_pools=ignore_pools,
            sync=False,
        )
        
    async def _symbol(self) -> str:
        symbol = await probe(self.address, ["symbol()(string)", "SYMBOL()(string)", "getSymbol()(string)"])
        if symbol is None:
            # Sometimes the above will fail if the symbol method returns bytes32, as with MKR. Let's try this.
            symbol = await probe(self.address, ["symbol()(bytes32)"])
            if symbol:
                symbol = hex_to_string(symbol)
        if symbol:
            return symbol
        # we've failed to fetch
        self.__raise_exception('symbol')
    
    async def _name(self) -> str:
        name = await probe(self.address, ["name()(string)", "NAME()(string)", "getName()(string)"])
        if name is None:
            # Sometimes the above will fail if the name method returns bytes32, as with MKR. Let's try this.
            name = await probe(self.address, ["name()(bytes32)"])
            if name:
                name = hex_to_string(name)
        if name:
            return name
        # we've failed to fetch
        self.__raise_exception('name')

    def __raise_exception(self, fn_name: str):
        raise NonStandardERC20(f'''
            Unable to fetch `{fn_name}` for {self.address} on {Network.printable()}
            If the contract is verified, please check to see if it has a strangely named
            `{fn_name}` method and create an issue on https://github.com/BobTheBuidler/ypricemagic
            with the contract address and correct method name so we can keep things going smoothly :)''') from None

    # These dundermethods are created by a_sync for the async_properties on this class
    __symbol__: HiddenMethodDescriptor[Self, str]
    __name__: HiddenMethodDescriptor[Self, str]
    __decimals__: HiddenMethodDescriptor[Self, int]
    __scale__: HiddenMethodDescriptor[Self, int]


@final
class WeiBalance(a_sync.ASyncGenericBase):
    # defaults are stored as class vars to keep instance dicts smaller
    block: Optional[Block] = None
    asynchronous: bool = False
    _skip_cache: bool = ENVS.SKIP_CACHE
    _ignore_pools: Tuple[Pool, ...] = ()
    def __init__(
        self, 
        balance: int,
        token: AnyAddressType,
        block: Optional[Block] = None,
        *,
        skip_cache: bool = ENVS.SKIP_CACHE,
        ignore_pools: Tuple[Pool, ...] = (),
        asynchronous: bool = False,
        ) -> None:

        if asynchronous != self.asynchronous:
            self.asynchronous = asynchronous
        self.balance = Decimal(balance)
        self.token = ERC20(str(token), asynchronous=self.asynchronous)
        if block != self.block:
            self.block = block
        super().__init__()
        if skip_cache != self._skip_cache:
            self._skip_cache = skip_cache
        if ignore_pools != self._ignore_pools:
            self._ignore_pools = ignore_pools

    def __str__(self) -> str:
        return str(self.balance)
    
    def __repr__(self) -> str:
        return f"<WeiBalance token={self.token} balance={self.balance} block={self.block}>"
    
    def __hash__(self) -> int:
        return hash((self.balance, self.token, self.block, self._skip_cache, self._ignore_pools))

    def __bool__(self) -> bool:
        return bool(self.balance)

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, int):
            return __o == self.balance
        elif isinstance(__o, WeiBalance):
            return (
                self.balance == __o.balance and 
                self.token == __o.token and 
                self.block == __o.block and 
                self._skip_cache == __o._skip_cache and
                self._ignore_pools == __o._ignore_pools
            )
        return False
    
    def __lt__(self, __o: object) -> bool:
        if isinstance(__o, int):
            return __o < self.balance
        elif isinstance(__o, WeiBalance):
            if self.token != __o.token:
                raise ValueError(f"'<' only supported between {self.__class__.__name__} instances when denominated in the same token.") from None
            return self.balance < __o.balance
        raise TypeError(f"'<' not supported between instances of '{self.__class__.__name__}' and '{__o.__class__.__name__}'") from None

    def __ge__(self, __o: object) -> bool:
        if __o < self:
            return True
        elif type(__o) is type(self):
            return self == __o
        raise TypeError(f"'>=' not supported between instances of '{self.__class__.__name__}' and '{__o.__class__.__name__}'") from None
    
    def __radd__(self, __o: Union["WeiBalance", Literal[0]]) -> "WeiBalance":
        if __o == 0:
            return self
        try:
            if self.token != __o.token:
                raise ValueError(f"addition not supported between instances of '{self.__class__.__name__}' and '{__o.__class__.__name__}'") from None
            if self.block != __o.block:
                raise ValueError("addition not supported between balances at different block heights") from None
            if self._skip_cache != __o._skip_cache:
                raise ValueError("addition not supported between balances with different `_skip_cache` values") from None
            if self._ignore_pools != __o._ignore_pools:
                raise ValueError("addition not supported between balances with different `_ignore_pools` values") from None
            return WeiBalance(self.balance - __o.balance, self.token, self.block, skip_cache=self._skip_cache, ignore_pools=self._ignore_pools)
        except AttributeError:
            raise TypeError(f"right addition not supported between instances of '{type(self).__name__}' and '{type(__o).__name__}'") from None
    
    def __add__(self, __o: "WeiBalance") -> "WeiBalance":
        try:
            if self.token != __o.token:
                raise ValueError(f"addition not supported between instances of '{self.__class__.__name__}' and '{__o.__class__.__name__}'") from None
            if self.block != __o.block:
                raise ValueError("addition not supported between balances at different block heights") from None
            if self._skip_cache != __o._skip_cache:
                raise ValueError("addition not supported between balances with different `_skip_cache` values") from None
            if self._ignore_pools != __o._ignore_pools:
                raise ValueError("addition not supported between balances with different `_ignore_pools` values") from None
            return WeiBalance(self.balance - __o.balance, self.token, self.block, skip_cache=self._skip_cache, ignore_pools=self._ignore_pools)
        except AttributeError:
            raise TypeError(f"addition not supported between instances of '{type(self).__name__}' and '{type(__o).__name__}'") from None
    
    def __sub__(self, __o: "WeiBalance") -> "WeiBalance":
        try:
            if self.token != __o.token:
                raise ValueError(f"subtraction not supported between instances of '{self.__class__.__name__}' and '{__o.__class__.__name__}'") from None
            if self.block != __o.block:
                raise ValueError("subtraction not supported between balances at different block heights") from None
            if self._skip_cache != __o._skip_cache:
                raise ValueError("subtraction not supported between balances with different `_skip_cache` values") from None
            if self._ignore_pools != __o._ignore_pools:
                raise ValueError("subtraction not supported between balances with different `_ignore_pools` values") from None
            return WeiBalance(self.balance - __o.balance, self.token, self.block, skip_cache=self._skip_cache, ignore_pools=self._ignore_pools)
        except AttributeError:
            raise TypeError(f"subtraction not supported between instances of '{type(self).__name__}' and '{type(__o).__name__}'") from None
    
    def __mul__(self, __o: Union[int, float, Decimal]) -> "WeiBalance":
        if not isinstance(__o, (int, float, Decimal)):
            raise TypeError(f"multiplication not supported between instances of '{type(self).__name__}' and '{type(__o).__name__}'") from None
        return WeiBalance(self.balance * Decimal(__o), self.token, self.block, skip_cache=self._skip_cache, ignore_pools=self._ignore_pools)
    
    def __truediv__(self, __o: Union[int, float, Decimal]) -> "WeiBalance":
        if not isinstance(__o, (int, float, Decimal)):
            raise TypeError(f"division not supported between instances of '{type(self).__name__}' and '{type(__o).__name__}'") from None
        return WeiBalance(self.balance / Decimal(__o), self.token, self.block, skip_cache=self._skip_cache, ignore_pools=self._ignore_pools)
    
    @a_sync.aka.property
    async def readable(self) -> Decimal:
        if self.balance == 0:
            return 0
        scale = await self.token.__scale__
        readable = self.balance / scale
        self._logger.debug("balance: %s  decimals: %s  readable: %s", self.balance, str(scale).count("0"), readable)
        return readable
    __readable__: HiddenMethodDescriptor[Self, Decimal]
    
    @a_sync.aka.property
    async def price(self) -> Decimal:
        price = Decimal(await self.token.price(block=self.block, skip_cache=self._skip_cache, ignore_pools=self._ignore_pools, sync=False))
        self._logger.debug("balance: %s  price: %s", self, price)
        return price
    __price__: HiddenMethodDescriptor[Self, Decimal]
    
    @a_sync.aka.property
    async def value_usd(self) -> Decimal:
        if self.balance == 0:
            return 0
        balance, price = await asyncio.gather(self.__readable__, self.__price__)
        value = balance * price
        self._logger.debug("balance: %s  price: %s  value: %s", balance, price, value)
        return value
    __value_usd__: HiddenMethodDescriptor[Self, Decimal]
    
    @cached_property
    def _logger(self) -> logging.logging.Logger:
        return logging.get_price_logger(self.token.address, self.block, extra=self.__class__.__name__)


class _Loader(ContractBase):
    """Used for use cases where you need to load data thru present time before proceeding, and then continue loading data in the background."""
    __slots__ = "_loaded", "_init_block", "__exc", "__task",
    def __init__(self, address: Address, asynchronous: bool = False):
        super().__init__(address, asynchronous=asynchronous)
        self._init_block = auto_retry(web3.eth.get_block_number)()
        self._loaded = None
        self.__exc = None
        self.__task = None
    def __await__(self) -> Generator[Any, None, Literal[True]]:
        """Returns `True` once the `_Loader` has loaded all relevant data thru the current block"""
        return self.loaded.__await__()
    @abc.abstractmethod
    async def _load(self) -> NoReturn:
        """
        `self._load` is the coro that will run in the daemon task associated with this _Loader.
        Your implementation MUST set Event `self._loaded` once data has been loaded thru the current block, or it will hang indefinitely.
        """
    @property
    def loaded(self) -> Awaitable[Literal[True]]:
        """Returns `True` once the `_Loader` has loaded all relevant data thru the current block"""
        self._task  # ensure task is running and not errd
        if self._loaded is None:
            self._loaded = a_sync.Event(name=self)
        return self._loaded.wait()
    @property
    def _task(self) -> "asyncio.Task[NoReturn]":
        """The task that runs `self._load() for this `_Loader`"""
        if self.__exc:
            # create a new duplicate exc instead of building a massive traceback on the original
            raise type(self.__exc)(*self.__exc.args).with_traceback(self.__tb)
        if self.__task is None:
            logger.debug("creating loader task for %s", self)
            self.__task = asyncio.create_task(coro=self.__load(), name=f"{self}.__load()")
            self.__task.add_done_callback(self._done_callback)
        return self.__task
    def _done_callback(self, task: "asyncio.Task[Any]") -> None:
        """called on `self._task` when it completes, if applicable"""
        if e := task.exception():
            logger.error("exception while loading %s: %s", self, e)
            logger.exception(e)
            self.__task = None
    async def __load(self) -> NoReturn:
        """loads the loader and catches any exceptions"""
        try:
            await self._load()
        except Exception as e:
            import traceback
            self.__exc = e
            self.__tb = e.__traceback__
            # no need to hold vars in memory
            traceback.clear_frames(self.__tb)
            raise


class _EventsLoader(_Loader):
    """Used for use cases where you need to load event data thru present time before proceeding, and then continue loading data in the background."""
    @abc.abstractproperty
    def _events(self) -> "Events":
        ...
    @property
    def loaded(self) -> Awaitable[Literal[True]]:
        self._task  # ensure task is running and not err'd
        if self._loaded is None:
            self._loaded = asyncio.ensure_future(self._events._lock.wait_for(self._init_block))
        return self._loaded
    async def _load(self) -> NoReturn:
        # TODO: extend this for constant loading
        async for _ in self._events.events(self._init_block):
            pass



 
import asyncio
import logging
from contextlib import suppress
from decimal import Decimal
from functools import cached_property
from typing import Any, AsyncIterator, Dict, List, Optional, Tuple

import a_sync
import a_sync.exceptions
import brownie
import dank_mids
from a_sync.a_sync import HiddenMethodDescriptor
from brownie import chain
from brownie.network.event import _EventItem
from dank_mids.exceptions import Revert
from multicall import Call
from typing_extensions import Self

from y import ENVIRONMENT_VARIABLES as ENVS
from y import convert
from y._decorators import continue_on_revert, stuck_coro_debugger
from y.classes.common import ERC20, ContractBase, WeiBalance
from y.constants import STABLECOINS, WRAPPED_GAS_COIN, sushi, usdc, weth
from y.contracts import Contract, contract_creation_block_async
from y.datatypes import (Address, AddressOrContract, AnyAddressType, Block,
                         Pool, UsdPrice)
from y.exceptions import (CantFindSwapPath, ContractNotVerified,
                          NonStandardERC20, NotAUniswapV2Pool, TokenNotFound,
                          call_reverted, continue_if_call_reverted)
from y.interfaces.uniswap.factoryv2 import UNIV2_FACTORY_ABI
from y.networks import Network
from y.prices import magic
from y.prices.dex.uniswap.v2_forks import (ROUTER_TO_FACTORY,
                                           ROUTER_TO_PROTOCOL, special_paths)
from y.utils.events import ProcessedEvents
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)

Path = List[AddressOrContract]
Reserves = Tuple[int,int,int]

factory_helper_address = {
    # put special case addresses here
}.get(chain.id, "0xE57Bfd650A7771E401d56d4b2CA22d9f8f51D3D9")

try:
    FACTORY_HELPER = Contract(factory_helper_address)
except ContractNotVerified:
    FACTORY_HELPER = None


class UniswapV2Pool(ERC20):
    """
    Represents a Uniswap V2 liquidity pool.
    """

    # default stored as class var to keep instance dicts smaller
    __types_assumed = True
    "True if we're assuming types based on normal univ2 abi, False if we checked via block explorer."

    __slots__ = 'get_reserves',
    def __init__(
        self, 
        address: AnyAddressType, 
        token0: Optional[Address] = None, 
        token1: Optional[Address] = None, 
        deploy_block: Optional[int] = None, 
        asynchronous: bool = False,
    ):
        super().__init__(address, asynchronous=asynchronous)
        self.get_reserves = Call(self.address, 'getReserves()((uint112,uint112,uint32))')
        if deploy_block:
            self._deploy_block = deploy_block
        if token0:
            self.token0 = token0
        if token1:
            self.token1 = token1
        
    @a_sync.aka.cached_property
    async def factory(self) -> Address:
        try: return await raw_call(self.address, 'factory()', output='address', sync=False)
        except ValueError as e:
            if call_reverted(e):
                raise NotAUniswapV2Pool(self) from e
            # `is not a valid ETH address` means we got some kind of response from the chain.
            # but couldn't convert to address. If it happens to be a goofy but
            # verified uni fork, maybe we can get factory this way
            okay_errors = ['is not a valid ETH address','invalid opcode','invalid jump destination']
            if all(msg not in str(e) for msg in okay_errors):
                raise
            contract = await Contract.coroutine(self.address)
            try: 
                return await contract.factory
            except AttributeError as exc:
                raise NotAUniswapV2Pool(self) from exc
    __factory__: HiddenMethodDescriptor[Self, Address]

    @a_sync.aka.property
    async def tokens(self) -> Tuple[ERC20, ERC20]:
        return await asyncio.gather(self.__token0__, self.__token1__)
    __tokens__: HiddenMethodDescriptor[Self, Tuple[ERC20, ERC20]]
    
    @a_sync.aka.cached_property
    async def token0(self) -> ERC20:
        try:
            if token0 := await Call(self.address, ['token0()(address)']):
                return ERC20(token0, asynchronous=self.asynchronous)
        except ValueError as e:
            continue_if_call_reverted(e)
        raise NotAUniswapV2Pool(self)
    __token0__: HiddenMethodDescriptor[Self, ERC20]

    @a_sync.aka.cached_property
    async def token1(self) -> ERC20:
        try:
            if token1 := await Call(self.address, ['token1()(address)']):
                return ERC20(token1, asynchronous=self.asynchronous)
        except ValueError as e:
            continue_if_call_reverted(e)
        raise NotAUniswapV2Pool(self)
    __token1__: HiddenMethodDescriptor[Self, ERC20]
    
    @a_sync.a_sync(ram_cache_ttl=ENVS.CACHE_TTL)
    @stuck_coro_debugger
    async def get_price(self, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> Optional[UsdPrice]:
        tvl = await self.tvl(block=block, skip_cache=skip_cache, sync=False)
        if tvl is not None:
            return UsdPrice(tvl / Decimal(await self.total_supply_readable(block=block, sync=False)))
    
    @a_sync.a_sync(ram_cache_maxsize=None, ram_cache_ttl=ENVS.CACHE_TTL)
    async def get_token_out(self, token_in: Address) -> ERC20:
        if token_in == (token0 := await self.__token0__):
            # these return instantly since theyre already cached
            return await self.__token1__
        elif token_in == (token1 := await self.__token1__):
            # these return instantly since theyre already cached
            return await self.__token0__
        raise TokenNotFound(token_in, [token0, token1]) from None
    
    @stuck_coro_debugger
    async def reserves(self, *, block: Optional[Block] = None) -> Optional[Tuple[WeiBalance, WeiBalance]]:
        try:
            reserves = await self.get_reserves.coroutine(block_id=block)
        except Exception as e:
            if not call_reverted(e):
                raise
            reserves = None

        if reserves is None and self.__types_assumed:
            try:
                await self._check_return_types()
            except AttributeError as e:
                raise NotAUniswapV2Pool(self) from e
            return await self.reserves(block=block, sync=False)
        
        if reserves is None and self._verified:
            # This shouldn't really run anymore, maybe delete
            contract = await Contract.coroutine(self.address)
            try:
                reserves = await contract.getReserves.coroutine(block_identifier=block)
                types = ",".join(output["type"] for output in contract.getReserves.abi["outputs"])
                logger.warning(f'abi for getReserves for {contract} is {types}')
            except Exception as e:
                if not call_reverted(e):
                    raise
                    
        if reserves is None:
            return None

        # NOTE: using `__token0__` and `__token1__` is faster than `__tokens__` since they're already cached and return instantly
        #       it also creates 2 fewer tasks and 1 fewer future than `__tokens__` since there is no use of `asyncio.gather`.
        return (
            WeiBalance(reserves[0], await self.__token0__, block=block),
            WeiBalance(reserves[1], await self.__token1__, block=block),
        )

    @stuck_coro_debugger
    async def tvl(self, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> Optional[Decimal]:
        # start these tasks now
        price_tasks: a_sync.TaskMapping[ERC20, UsdPrice]
        price_tasks = ERC20.price.map(
            # NOTE: using `__token0__` and `__token1__` is faster than `__tokens__` since they're already cached and return instantly
            #       it also creates 2 fewer tasks and 1 fewer future than `__tokens__` since there is no use of `asyncio.gather`.
            [await self.__token0__, await self.__token1__], 
            block=block, 
            return_None_on_failure=True, 
            skip_cache=skip_cache,
        )

        reserves: Tuple[WeiBalance, WeiBalance]
        if (reserves := await self.reserves(block=block, sync=False)) is None:
            await price_tasks.close()
            return None

        prices = await price_tasks.values(pop=True)
        if vals := [
            Decimal(await reserve.__readable__) * Decimal(price)
            for reserve, price in zip(reserves, prices)
            if price is not None
        ]:
            if len(vals) == 1:
                vals *= 2
            if len(vals) == 2:
                logger.debug('reserves: %s', reserves)
                logger.debug('prices: %s', prices)
                logger.debug('vals: %s', vals)
                return sum(vals)
            else:
                raise Exception("how did we get here?") from None
    
    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=100_000, ram_cache_ttl=60*60, semaphore=10_000)  # lets try a semaphore here
    async def check_liquidity(self, token: Address, block: Block) -> int:
        """
        Check the liquidity of a specific token in the pool at a given block.

        Args:
            token: The address of the token to check.
            block: The block number to query.

        Returns:
            The liquidity of the token in the pool.

        Raises:
            TokenNotFound: If the token is not one of the two tokens in the liquidity pool.
        """
        logger.debug("checking %s liquidity for %s at %s", self, token, block)
        if block and block < await self.deploy_block(sync=False):
            logger.debug("block %s is before %s deploy block", block, self)
            return 0
        if reserves := await self.reserves(block=block, sync=False):
            balance: WeiBalance
            for balance in reserves:
                if token == balance.token:
                    liquidity = balance.balance
                    logger.debug("%s liquidity for %s at %s is %s", self, token, block, liquidity)
                    return liquidity
            raise TokenNotFound(token, reserves)
        return 0

    @stuck_coro_debugger
    async def is_uniswap_pool(self, block: Optional[Block] = None) -> bool:
        """
        Check if this contract is a valid liquidity pool for Uniswap V2 or one of its forks.

        Args:
            block (optional): The block number to query. Defaults to latest block.

        Returns:
            True if the contract is a valid Uniswap V2 pool, False otherwise.
        """
        try:
            return all(await asyncio.gather(self.reserves(block=block, sync=False), self.total_supply(block, sync=False)))
        except NotAUniswapV2Pool:
            return False
    
    async def _check_return_types(self) -> None:
        if not self.__types_assumed:
            return
        try:
            contract = await Contract.coroutine(self.address)
            reserves_types = ",".join(output["type"] for output in contract.getReserves.abi["outputs"])
            self._verified = True
            assert reserves_types.count(',') == 2, reserves_types
            self.get_reserves = Call(self.address, f'getReserves()(({reserves_types}))')
        except ContractNotVerified:
            self._verified = False
        self.__types_assumed = False


class PoolsFromEvents(ProcessedEvents[UniswapV2Pool]):
    PairCreated = "0x0d3648bd0f6ba80134a33ba9275ac585d9d315f0ad8355cddefde31afa28d0e9"
    __slots__ = "asynchronous", "label"
    def __init__(self, factory: AnyAddressType, label: str, asynchronous: bool = False):
        self.asynchronous = asynchronous
        self.label = label
        super().__init__(addresses=[factory], topics=[[self.PairCreated]])
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} label={self.label}>"
    def pools(self, to_block: Optional[int] = None) -> AsyncIterator[UniswapV2Pool]:
        return self._objects_thru(block=to_block)
    def _get_block_for_obj(self, obj: UniswapV2Pool) -> int:
        return obj._deploy_block
    def _process_event(self, event: _EventItem) -> UniswapV2Pool:
        pool = UniswapV2Pool(
            address=event["pair"], 
            token0=event["token0"], 
            token1=event["token1"], 
            asynchronous=self.asynchronous,
        )
        # Do this here instead of in the init in case the user inited their own UniswapV2Pool object previoulsy, which is now the singleton
        pool._deploy_block = event.block_number
        return pool
    

class UniswapRouterV2(ContractBase):
    def __init__(self, router_address: AnyAddressType, *args: Any, **kwargs: Any) -> None:
        super().__init__(router_address, *args, **kwargs)

        self.label = ROUTER_TO_PROTOCOL[self.address]
        self.factory = ROUTER_TO_FACTORY[self.address]
        self.special_paths = special_paths(self.address)
        self.get_amounts_out = Call(self.address, "getAmountsOut(uint,address[])(uint[])")

        # we need the factory contract object cached in brownie so we can decode logs properly
        if not ContractBase(self.factory, asynchronous=self.asynchronous)._is_cached:
            brownie.Contract.from_abi('UniClone Factory [forced]', self.factory, UNIV2_FACTORY_ABI)

        self._events = PoolsFromEvents(self.factory, self.label, asynchronous=self.asynchronous)
    
    def __str__(self) -> str:
        return self.__repr__()
    
    def __repr__(self) -> str:
        return f"<UniswapV2Router {self.label} '{self.address}'>"

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=500)
    async def get_price(
        self,
        token_in: Address,
        block: Optional[Block] = None,
        token_out: Address = usdc.address,
        paired_against: Address = WRAPPED_GAS_COIN,
        skip_cache: bool = ENVS.SKIP_CACHE,
        ignore_pools: Tuple[Pool, ...] = (),
        ) -> Optional[UsdPrice]:
        """
        Calculate a price based on Uniswap Router quote for selling one `token_in`.
        Always uses intermediate WETH pair if `[token_in,weth,token_out]` swap path available.
        """

        token_in, token_out, path = str(token_in), str(token_out), None

        if chain.id == Network.BinanceSmartChain and token_out == usdc.address:
            busd = await Contract.coroutine("0xe9e7CEA3DedcA5984780Bafc599bD69ADd087D56")
            token_out = busd.address

        if token_in in STABLECOINS:
            return 1

        try:
            amount_in = await ERC20(token_in, asynchronous=True).scale
        except NonStandardERC20:
            return None

        if token_in in [weth.address, WRAPPED_GAS_COIN] and token_out in STABLECOINS:
            path = [token_in, token_out]

        elif str(token_out) in STABLECOINS:
            with suppress(CantFindSwapPath):
                path = await self.get_path_to_stables(token_in, block, _ignore_pools=ignore_pools, sync=False)
                logger.debug('smrt')

        # If we can't find a good path to stables, we might still be able to determine price from price of paired token
        if path is None and (deepest_pool:= await self.deepest_pool(token_in, block, _ignore_pools=ignore_pools, sync=False)):
            logger.debug('deepest pool: %s', deepest_pool)
            paired_with = await deepest_pool.get_token_out(token_in, sync=False)
            path = [token_in, paired_with]
            quote, out_scale = await asyncio.gather(self.get_quote(amount_in, path, block=block, sync=False), ERC20(path[-1], asynchronous=True).scale)
            logger.debug('quote: %s', quote)
            if quote is not None:
                amount_out = Decimal(quote[-1]) / out_scale  
                fees = Decimal(0.997) ** (len(path) - 1)
                amount_out /= fees
                paired_with_price = await magic.get_price(paired_with, block, fail_to_None=True, skip_cache=skip_cache, ignore_pools=(*ignore_pools, deepest_pool), sync=False)

                if paired_with_price:
                    return amount_out * Decimal(paired_with_price)

        # If we still don't have a workable path, try this smol brain method
        if path is None:
            path = self._smol_brain_path_selector(token_in, token_out, paired_against)
            # NOTE: does this ever run anymore? can we take it out?
            logger.warning('using smol brain path selector')

        fees = 0.997 ** (len(path) - 1)
        logger.debug('router: %s     path: %s', self.label, path)
        quote, out_scale = await asyncio.gather(self.get_quote(amount_in, path, block=block, sync=False), ERC20(path[-1],asynchronous=True).scale)
        if quote is not None:
            amount_out = quote[-1] / out_scale
            return UsdPrice(amount_out / fees)


    @continue_on_revert
    @stuck_coro_debugger
    async def get_quote(self, amount_in: int, path: Path, block: Optional[Block] = None) -> Tuple[int,int]:
        if not self._is_cached:
            return await self.get_amounts_out.coroutine((amount_in, path), block_id=block)
        try:
            return await self.contract.getAmountsOut.coroutine(amount_in, path, block_identifier=block)
        # TODO figure out how to best handle uni forks with slight modifications.
        # Sometimes the below "else" code will not work with modified methods. Brownie works for now.
        except Exception as e:
            strings = [
                "INSUFFICIENT_INPUT_AMOUNT",
                "INSUFFICIENT_LIQUIDITY",
                "INSUFFICIENT_OUT_LIQUIDITY",
                "Sequence has incorrect length",
            ]
            if not call_reverted(e) and all(s not in str(e) for s in strings):
                raise
    
    @a_sync.aka.cached_property
    @stuck_coro_debugger
    async def pools(self) -> List[UniswapV2Pool]:
        logger.info('Fetching pools for %s on %s. If this is your first time using ypricemagic, this can take a while. Please wait patiently...', self.label, Network.printable())
        pools = [pool async for pool in self._events.pools(to_block=await dank_mids.eth.block_number)]
        all_pairs_len = await raw_call(self.factory, 'allPairsLength()', output='int', sync=False)
        if len(pools) > all_pairs_len:
            raise NotImplementedError('this shouldnt happen again')
        elif (to_get := all_pairs_len - len(pools)): # <
            logger.debug("Oh no! Looks like your node can't look back that far. Checking for the missing %s pools...", to_get)
            factory = await Contract.coroutine(self.factory)
            for i, pool_address in enumerate(await factory.allPairs.map(range(to_get))):
                pool = UniswapV2Pool(address=pool_address, asynchronous=self.asynchronous)
                pools.insert(i, pool)
            logger.debug('Done fetching %s missing pools on %s', to_get, self.label)
        token0s = UniswapV2Pool.token0.map(pools).values(pop=True)
        token1s = UniswapV2Pool.token1.map(pools).values(pop=True)
        tokens = set(await token0s + await token1s)
        logger.info('Loaded %s pools supporting %s tokens on %s', len(pools), len(tokens), self.label)
        return pools
    __pools__: HiddenMethodDescriptor[Self, List[UniswapV2Pool]]

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=None, ram_cache_ttl=60*60)
    async def get_pools_for(self, token_in: Address, block: Optional[Block] = None) -> Dict[UniswapV2Pool, Address]:
        if self._supports_uniswap_helper:
            try:
                pools = [
                    UniswapV2Pool(pool, asynchronous=self.asynchronous)
                    for pool in await FACTORY_HELPER.getPairsFor.coroutine(self.factory, token_in, block_identifier=block)
                ]
            except Exception as e:
                if not (call_reverted(e) or "out of gas" in str(e) or "timeout" in str(e)):
                    raise
                pools = await self.__pools__

        else:
            pools = await self.__pools__

        pool_to_token_out = {}
        for pool in pools:
            # these will return immediately since the pools are already loaded by this point
            if token_in == await pool.__token0__:
                pool_to_token_out[pool] = await pool.__token1__
            elif token_in == await pool.__token1__:
                pool_to_token_out[pool] = await pool.__token0__
        return pool_to_token_out

    @stuck_coro_debugger
    async def pools_for_token(self, token_address: Address, block: Optional[Block] = None, _ignore_pools: Tuple[UniswapV2Pool,...] = ()) -> AsyncIterator[UniswapV2Pool]:
        pools: Dict[UniswapV2Pool, Address]
        
        if chain.id == Network.Mainnet and token_address == WRAPPED_GAS_COIN and self.label == "uniswap v2":
            # This will run out of gas if we use the helper so we bypass it with a known liquid pool
            pools = {UniswapV2Pool("0xB4e16d0168e52d35CaCD2c6185b44281Ec28C9Dc", asynchronous=True): "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48"}
        else:
            try:
                pools = await self.get_pools_for(token_address, sync=False)
            except Exception as e:
                raise
                if 'out of gas' not in str(e) and not call_reverted(e):
                    e.args = (*e.args, self, token_address, block)
                    raise
                try:
                    # if it fails with no block we will try once with a block before we fetch the long way
                    pools = await self.get_pools_for(token_address, block=block, sync=False)
                except Exception as e:
                    e.args = (*e.args, self, token_address, block)
                    raise
        for pool in _ignore_pools:
            pools.pop(pool, None)
        if not pools:
            return
        elif block is None:
            for pool in pools:
                yield pool
        else:
            async for pool, deploy_block in ERC20.deploy_block.map(when_no_history_return_0=True).map(pools):
                if deploy_block <= block:
                    yield pool

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=500)
    async def deepest_pool(self, token_address: AnyAddressType, block: Optional[Block] = None, _ignore_pools: Tuple[UniswapV2Pool,...] = ()) -> Optional[UniswapV2Pool]:
        """returns the deepest pool for `token_address` at `block`, excluding pools in `_ignore_pools`"""
        token_address = convert.to_address(token_address)
        if token_address == WRAPPED_GAS_COIN or token_address in STABLECOINS:
            return await self.deepest_stable_pool(token_address, block, sync=False)
        if self._supports_uniswap_helper and (block is None or block >= await contract_creation_block_async(FACTORY_HELPER)):
            try:
                deepest_pool, deepest_pool_depth = await self.deepest_pool_for(token_address, block, ignore_pools=_ignore_pools)
                return None if deepest_pool == brownie.ZERO_ADDRESS else UniswapV2Pool(deepest_pool, asynchronous=self.asynchronous)
            except Revert as e:
                # TODO: debug me!
                logger.debug('helper reverted for %s at block %s ignore_pools %s: %s', token_address, block, _ignore_pools, e)
            except ValueError as e:
                if "out of gas" not in str(e):
                    raise
                logger.debug('helper out of gas for %s at block %s ignore_pools %s: %s', token_address, block, _ignore_pools, e)

        pools = self.pools_for_token(token_address, block, _ignore_pools=_ignore_pools)
        logger.debug("checking %s liquidity at block %s for pools %s", token_address, block, pools)

        deepest_pool: UniswapV2Pool
        async for deepest_pool in UniswapV2Pool.check_liquidity.map(pools, token=token_address, block=block).keys(pop=True).aiterbyvalues(reverse=True):
            return deepest_pool

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=500)
    async def deepest_stable_pool(self, token_address: AnyAddressType, block: Optional[Block] = None, _ignore_pools: Tuple[UniswapV2Pool,...] = ()) -> Optional[UniswapV2Pool]:
        """returns the deepest pool for `token_address` at `block` which has `token_address` paired with a stablecoin, excluding pools in `_ignore_pools`"""
        token_out_tasks: a_sync.TaskMapping[UniswapV2Pool, ERC20]
        deepest_stable_pool: UniswapV2Pool

        pools = self.pools_for_token(convert.to_address(token_address), block=block, _ignore_pools=_ignore_pools)
        token_out_tasks = UniswapV2Pool.get_token_out.map(token_in=token_address)
        if stable_pools := [pool async for pool, paired_with in token_out_tasks.map(pools) if paired_with in STABLECOINS]:
            del token_out_tasks

            if self._supports_uniswap_helper and (block is None or block >= await contract_creation_block_async(FACTORY_HELPER, when_no_history_return_0=True)):
                deepest_stable_pool, deepest_stable_pool_balance = await FACTORY_HELPER.deepestPoolForFrom.coroutine(token_address, stable_pools, block_identifier=block)
                return None if deepest_stable_pool == brownie.ZERO_ADDRESS else UniswapV2Pool(deepest_stable_pool, asynchronous=self.asynchronous)
        
            async for deepest_stable_pool, depth in UniswapV2Pool.check_liquidity.map(stable_pools, token=token_address, block=block).items(pop=True).aiterbyvalues(reverse=True):
                return deepest_stable_pool

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=500)
    async def get_path_to_stables(self, token: AnyAddressType, block: Optional[Block] = None, _loop_count: int = 0, _ignore_pools: Tuple[UniswapV2Pool,...] = ()) -> Path:
        if _loop_count > 10:
            raise CantFindSwapPath
        token_address = convert.to_address(token)
        path = [token_address]
        deepest_pool = await self.deepest_pool(token_address, block, _ignore_pools, sync=False)
        if deepest_pool:
            paired_with = await deepest_pool.get_token_out(token_address, sync=False)
            from y.prices.utils.buckets import check_bucket
            if await check_bucket(paired_with, sync=False) and _loop_count == 0:
                # let's just use the other token to get the price
                return None
            deepest_stable_pool = await self.deepest_stable_pool(token_address, block, _ignore_pools=_ignore_pools, sync=False)
            if deepest_stable_pool and deepest_pool == deepest_stable_pool:
                path.append(await deepest_stable_pool.get_token_out(token_address, sync=False))
                return path

            if path == [token_address]:
                with suppress(CantFindSwapPath):
                    path.extend(
                        await self.get_path_to_stables(
                            paired_with,
                            block=block, 
                            _loop_count=_loop_count+1, 
                            _ignore_pools=tuple(list(_ignore_pools) + [deepest_pool]),
                            sync=False,
                        )
                    )

        if path == [token_address]:
            raise CantFindSwapPath(f'Unable to find swap path for {token_address} on {Network.printable()}')

        return path

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=100_000, ram_cache_ttl=60*60)
    async def check_liquidity(self, token: Address, block: Block, ignore_pools = []) -> int:
        logger.debug("checking %s liquidity for %s at %s", self, token, block)
        if block and block < await contract_creation_block_async(self.factory):
            logger.debug("block %s is before %s deploy block")
            return 0
        if self._supports_uniswap_helper and (block is None or block >= await contract_creation_block_async(FACTORY_HELPER)):
            try:
                deepest_pool, liquidity = await self.deepest_pool_for(token, block, ignore_pools=ignore_pools)
                logger.debug("%s liquidity for %s at %s is %s", self, token, block, liquidity)
                return liquidity
            except Revert as e:
                # TODO: debug me!
                logger.debug('helper reverted on check_liquidity for %s at block %s: %s',token, block, e)
            except ValueError as e:
                if "out of gas" not in str(e):
                    raise
                logger.debug('helper out of gas on check_liquidity for %s at block %s: %s',token, block, e)
                
        pools = self.pools_for_token(token, block=block, _ignore_pools=ignore_pools)
        try:
            liquidity = await UniswapV2Pool.check_liquidity.max(pools, token=token, block=block, sync=False)
        except a_sync.exceptions.EmptySequenceError:
            liquidity = 0
        logger.debug("%s liquidity for %s at %s is %s", self, token, block, liquidity)
        return liquidity

    @a_sync.a_sync(ram_cache_maxsize=100_000, ram_cache_ttl=60*60)
    @stuck_coro_debugger
    async def deepest_pool_for(self, token: Address, block: Block = None, *, ignore_pools = []) -> Tuple[Address, int]:
        # sourcery skip: default-mutable-arg
        try:
            deepest = await FACTORY_HELPER.deepestPoolFor.coroutine(self.factory, token, ignore_pools, block_identifier=block)
            logger.debug("got deepest pool for %s at %s: %s from helper", token, block, deepest)
            return deepest
        except Exception as e:
            e.args = (*e.args, self.__repr__(), token, block, ignore_pools)
            raise
    
    @cached_property
    def _supports_uniswap_helper(self) -> bool:
        """returns True if our uniswap helper contract is supported, False if not"""
        return chain.id != Network.Mainnet and FACTORY_HELPER and self.label and self.label not in {"zipswap"}


    def _smol_brain_path_selector(self, token_in: AddressOrContract, token_out: AddressOrContract, paired_against: AddressOrContract) -> Path:
        # sourcery skip: assign-if-exp, lift-return-into-if, merge-duplicate-blocks, merge-else-if-into-elif, remove-redundant-if, remove-unnecessary-cast
        '''Chooses swap path to use for quote'''
        # NOTE: can we just delete this now? probably, must test
        token_in, token_out, paired_against = str(token_in), str(token_out), str(paired_against)

        if str(paired_against) in STABLECOINS and str(token_out) in STABLECOINS:            path = [token_in, paired_against]
        elif weth in (token_in, token_out):                                                 path = [token_in, token_out]
        elif sushi and paired_against == sushi and token_out != sushi:                      path = [token_in,sushi,weth,token_out]
        elif str(token_in) in self.special_paths and str(token_out) in STABLECOINS:         path = self.special_paths[str(token_in)]

        elif chain.id == Network.BinanceSmartChain:
            from y.constants import cake, wbnb
            if WRAPPED_GAS_COIN in (token_in, token_out):                                   path = [token_in, token_out]
            elif cake.address in (token_in, token_out):                                     path = [token_in, token_out]
            else:                                                                           path = [token_in,wbnb.address,token_out]
        else:
            if WRAPPED_GAS_COIN in (token_in, token_out):                                   path = [token_in, token_out]
            else:                                                                           path = [token_in, WRAPPED_GAS_COIN, token_out]

        return path
    




from enum import IntEnum
from typing import Optional

from brownie import chain


class Network(IntEnum):
    """
    A lightweight enum that enables lookup of chainids for popular blockchain networks.

    Each network is associated with its unique integer Chain ID.
    """

    # NOTE: If a particular network has changed names, do not remove
    # the old spec, simply add the new spec directly above the old one.
    # Both will work but Network(chainid) will return the new name. """

    Mainnet = 1
    """The Chain ID for Ethereum Mainnet"""

    Optimism = 10
    """The Chain ID for Optimism"""

    Cronos = 25
    """The Chain ID for Cronos Mainnet"""

    BinanceSmartChain = 56
    """The Chain ID for Binance Smart Chain"""

    OKEx = 66
    """The Chain ID for OKEx Chain"""

    Gnosis = 100
    """The Chain ID for xDai Chain (now Gnosis Chain)"""

    xDai = 100
    """The Chain ID for xDai Chain (now Gnosis Chain)"""

    Heco = 128
    """The Chain ID for Heco"""

    Polygon = 137
    """The Chain ID for Polygon (formerly Matic) Network"""

    Fantom = 250
    """The Chain ID for Fantom Opera Network"""

    Moonriver = 1285
    """The Chain ID for Moonriver Network"""

    Base = 8453
    """The Chain ID for Base"""

    Arbitrum = 42161
    """The Chain ID for Arbitrum One"""

    Avalanche = 43114
    """The Chain ID for Avalanche C-Chain"""

    Harmony = 1666600000
    """The Chain ID for Harmony Mainnet Shard 0"""

    Aurora = 1313161554
    """The Chain ID for Aurora"""

    @staticmethod
    def label(chain_id: Optional[int] = None) -> str:
        if not chain_id: chain_id = chain.id

        if chain_id == Network.Mainnet:                 return 'ETH'
        elif chain_id == Network.BinanceSmartChain:     return 'BSC'
        elif chain_id == Network.Gnosis:                return 'GNO'
        elif chain_id == Network.Heco:                  return 'HECO'
        elif chain_id == Network.Polygon:               return 'POLY'
        elif chain_id == Network.Fantom:                return 'FTM'
        elif chain_id == Network.Moonriver:             return 'MOVR'
        elif chain_id == Network.Arbitrum:              return 'ARB'
        elif chain_id == Network.Avalanche:             return 'AVAX'
        elif chain_id == Network.Harmony:               return 'ONE'
        elif chain_id == Network.Aurora:                return 'AURORA'
        elif chain_id == Network.OKEx:                  return 'OKEX'
        elif chain_id == Network.Cronos:                return 'CRO'
        elif chain_id == Network.Optimism:              return 'OPTI'
        elif chain_id == Network.Base:                  return 'BASE'
    
    @staticmethod
    def name(chain_id: Optional[int] = None) -> str:
        if not chain_id: chain_id = chain.id

        if chain_id == Network.Mainnet:                 return 'Mainnet'
        elif chain_id == Network.BinanceSmartChain:     return 'Binance Smart Chain'
        elif chain_id == Network.Gnosis:                return 'Gnosis'
        elif chain_id == Network.Heco:                  return 'Heco'
        elif chain_id == Network.Polygon:               return 'Polygon'
        elif chain_id == Network.Fantom:                return 'Fantom'
        elif chain_id == Network.Moonriver:             return 'Moonriver'
        elif chain_id == Network.Arbitrum:              return 'Arbitrum'
        elif chain_id == Network.Avalanche:             return 'Avalanche'
        elif chain_id == Network.Harmony:               return "HarmonyOne Shard 0"
        elif chain_id == Network.Aurora:                return 'Aurora'
        elif chain_id == Network.OKEx:                  return 'OKEx'
        elif chain_id == Network.Cronos:                return 'Cronos'
        elif chain_id == Network.Optimism:              return "Optimism"
        elif chain_id == Network.Base:                  return "Base"

    @staticmethod
    def printable(chain_id: Optional[int] = None) -> str:
        # will always work to print a readable string that identifies the network, even if network not supported
        if chain_id is None: chain_id = chain.id
        return Network.name(chain_id) or f"chain {chain_id}"




import logging
from datetime import datetime, timezone
from decimal import Decimal
from typing import Any, Iterable
from pony.orm import Database, DatabaseError, commit

from y._db import entities
from y._db.decorators import a_sync_write_db_session, retry_locked


logger = logging.getLogger(__name__)

class SQLError(ValueError):
    """
    Custom exception for SQL-related errors.
    
    This exception wraps an Exception raised when executing a SQL statement.
    It includes both the original error and the SQL statement that caused it.
    """

@retry_locked
def execute(sql: str, *, db: Database = entities.db) -> None:
    """
    Execute a SQL statement with retry logic for locked databases.

    This function attempts to execute the given SQL statement and commit the changes.
    If the database is locked, the operation will be retried based on the @:class:`retry_locked` decorator.
    
    Args:
        sql: The SQL statement to execute.
        db: The database to execute the statement on. Defaults to entities.db.

    Raises:
        SQLError: If there's an error executing the SQL statement, except for "database is locked" errors.

    Note:
        - The function logs the SQL statement at debug level before execution.
        - If a "database is locked" error occurs, it's re-raised to trigger the retry mechanism.
        - For all other DatabaseErrors, it logs a warning and raises a SQLError with the original error and SQL statement.
    """
    try:
        logger.debug("EXECUTING SQL")
        logger.debug(sql)
        db.execute(sql)
        commit()
    except DatabaseError as e:
        if str(e) == "database is locked":
            raise
        logger.warning("%s %s when executing SQL`%s`", e.__class__.__name__, e, sql)
        raise SQLError(e, sql) from e

def stringify_column_value(value: Any, provider: str) -> str:
    """
    Convert a Python value to a string representation suitable for SQL insertion.

    This function handles various Python types and converts them to SQL-compatible string representations.
    It supports different database providers (currently 'postgres' and 'sqlite') for certain data types.

    Args:
        value: The value to stringify. Can be None, bytes, str, int, Decimal, or datetime.
        provider: The database provider. Currently supports 'postgres' and 'sqlite'.

    Returns:
        A string representation of the value suitable for SQL insertion.

    Raises:
        NotImplementedError: If the value type is not supported or if an unsupported provider is specified for bytes.

    Note:
        - None values are converted to 'null'.
        - Bytes are handled differently for 'postgres' (converted to bytea) and 'sqlite' (converted to hex).
        - Strings are wrapped in single quotes.
        - Integers and Decimals are converted to their string representation.
        - Datetimes are converted to UTC and formatted as ISO8601 strings.
    """
    if value is None:
        return 'null'
    elif isinstance(value, bytes):
        if provider == 'postgres':
            #return f"E'\\x{value.hex()}"
            return f"'{value.decode()}'::bytea"
        elif provider == 'sqlite':
            return f"X'{value.hex()}'"
        raise NotImplementedError(provider)
    elif isinstance(value, str):
        return f"'{value}'"
    elif isinstance(value, (int, Decimal)):
        return str(value)
    elif isinstance(value, datetime):
        return f"'{value.astimezone(timezone.utc).isoformat()}'"
    else:
        raise NotImplementedError(type(value), value)

def build_row(row: Iterable[Any], provider: str) -> str:
    """
    Build a SQL row string from an iterable of values.

    This function takes an iterable of values and converts each value to its SQL string representation,
    then combines them into a single row string suitable for SQL insertion.

    Args:
        row: An iterable of values to be converted into a SQL row string.
        provider: The database provider to use for value conversion.

    Returns:
        A string representing a SQL row, formatted as (value1,value2,...).

    Note:
        This function uses :func:`stringify_column_value` internally to convert each value.
    """
    return f"({','.join(stringify_column_value(col, provider) for col in row)})"

@a_sync_write_db_session
def insert(
    entity_type: entities.db.Entity, 
    columns: Iterable[str], 
    items: Iterable[Iterable[Any]],
    *,
    db: Database = entities.db,
) -> None:
    """
    Perform a bulk insert operation into the database.

    This function constructs and executes an INSERT statement for multiple rows of data.
    It supports different syntax for SQLite and PostgreSQL databases.

    Args:
        entity_type: The database entity type to insert into.
        columns: An iterable of column names.
        items: An iterable of iterables, where each inner iterable represents a row of data to insert.
        db: The database to perform the insertion on. Defaults to entities.db.

    Raises:
        NotImplementedError: If an unsupported database provider is used.

    Note:
        - For SQLite, it uses 'INSERT OR IGNORE' syntax.
        - For PostgreSQL, it uses 'INSERT ... ON CONFLICT DO NOTHING' syntax.
    """
    entity_name = entity_type.__name__.lower()
    data = ",".join(build_row(i, db.provider_name) for i in items)
    if db.provider_name == 'sqlite':
        execute(f'insert or ignore into {entity_name} ({",".join(columns)}) values {data}', db=db)
    elif db.provider_name == 'postgres':
        execute(f'insert into {entity_name} ({",".join(columns)}) values {data} on conflict do nothing', db=db)
    else:
        raise NotImplementedError(db.provider_name)
    logger.debug('inserted %s %ss to ydb', len(items), entity_name)




import logging
from datetime import datetime, timezone
from dateutil import parser
from functools import lru_cache
from typing import Dict, Optional, Set

import a_sync
from pony.orm import commit, select
from brownie import chain

from y._db.decorators import a_sync_read_db_session, a_sync_write_db_session, a_sync_write_db_session_cached, log_result_count
from y._db.entities import Block, BlockAtTimestamp, Chain, insert

logger = logging.getLogger(__name__)


@a_sync_read_db_session
def get_chain() -> Chain:
    return Chain.get(id=chain.id) or insert(type=Chain, id=chain.id) or Chain[chain.id]

@lru_cache
def ensure_chain() -> None:
    """ensures that the chain object for the connected chain has been inserted to the db"""
    get_chain(sync=True)
    commit()

@a_sync_read_db_session
def get_block(number: int) -> Block:
    if block := Block.get(chain=chain.id, number=number):
        return block
    return insert(type=Block, chain=chain.id, number=number) or get_block(number, sync=True)

@a_sync_write_db_session_cached
def ensure_block(number: int) -> None:
    if number not in known_blocks():
        from y._db.utils._ep import _get_get_block
        get_block = _get_get_block()
        get_block(number, sync=True)

@a_sync_read_db_session
def get_block_timestamp(number: int) -> Optional[int]:
    if (ts := known_block_timestamps().pop(number, None)) is None:
        from y._db.utils._ep import _get_get_block
        get_block = _get_get_block()
        ts = get_block(number, sync=True).timestamp
    if ts:
        # some db providers return a string here, we must parse it
        if isinstance(ts, str):
            ts = parser.parse(ts)
        unix = ts.timestamp()
        logger.debug("got Block[%s, %s].timestamp from cache: %s, %s", chain.id, number, unix, ts)
        return unix

@a_sync_read_db_session
def get_block_at_timestamp(timestamp: datetime) -> Optional[int]:
    if block := known_blocks_for_timestamps().pop(timestamp, None):
        logger.debug("found block %s for %s in ydb", block, timestamp)
        return block
    elif entity := BlockAtTimestamp.get(chainid=chain.id, timestamp=timestamp):
        block = entity.block
        logger.debug("found block %s for %s in ydb", block, timestamp)
        return block

@a_sync_write_db_session
def _set_block_timestamp(block: int, timestamp: int) -> None:
    from y._db.utils._ep import _get_get_block
    get_block = _get_get_block()
    block = get_block(block, sync=True)
    timestamp = datetime.fromtimestamp(timestamp, tz=timezone.utc)
    block.timestamp = timestamp
    logger.debug("cached %s.timestamp %s", block, timestamp)

set_block_timestamp = a_sync.ProcessingQueue(_set_block_timestamp, num_workers=10, return_data=False)

@a_sync_write_db_session
def _set_block_at_timestamp(timestamp: datetime, block: int) -> None:
    insert(BlockAtTimestamp, chainid=chain.id, timestamp=timestamp, block=block)
    logger.debug("inserted block %s for %s", block, timestamp)

set_block_at_timestamp = a_sync.ProcessingQueue(_set_block_at_timestamp, num_workers=10, return_data=False)

# startup caches

@lru_cache(maxsize=1)
@log_result_count("blocks")
def known_blocks() -> Set[int]:
    """cache and return all known blocks for this chain to minimize db reads"""
    return set(select(b.number for b in Block if b.chain.id == chain.id))

@lru_cache(maxsize=1)
@log_result_count("block timestamps")
def known_block_timestamps() -> Dict[int, datetime]:
    """cache and return all known block timestamps for this chain to minimize db reads"""
    query = select((b.number, b.timestamp) for b in Block if b.chain.id == chain.id and b.timestamp)
    page_size = 100_000
    timestamps = {}
    for i in range((query.count() // page_size) + 1):
        for block, timestamp in query.page(i+1, page_size):
            timestamps[block] = timestamp
    return timestamps

@lru_cache(maxsize=1)
@log_result_count("blocks for timestamps")
def known_blocks_for_timestamps() -> Dict[datetime, int]:
    """return all known blocks for timestamps for this chain to minimize db reads"""
    return dict(select((x.timestamp, x.block) for x in BlockAtTimestamp if x.chainid == chain.id))




import functools
import logging
import os
from inspect import iscoroutinefunction

import a_sync
import eth_retry

from a_sync._typing import AnyFn, P, T
from a_sync.a_sync.function import ASyncDecorator
from brownie import chain
from joblib import Memory

from y import ENVIRONMENT_VARIABLES as ENVS


@eth_retry.auto_retry
def _memory():
    """
    Create and return a :class:`Memory` object for caching values for the currently connected blockchain.

    Returns:
        A :class:`~Memory` object configured with the current chain's cache directory.
    """
    return Memory(f"cache/{chain.id}", verbose=0)

memory = _memory()

a_sync_ttl_cache: ASyncDecorator = a_sync.a_sync(ram_cache_ttl=ENVS.CACHE_TTL)


# NOTE: we have an optional disk cache that I made for debugging and decided to keep in for my own convenience.
# TODO: make a real disk cache

try:

    import toolcache

    """User doesn't has toolcache, this diskcache decorator will work."""

    logger = logging.getLogger(__name__)
    cache_base_path = f'./cache/{chain.id}/'

    def optional_async_diskcache(fn: AnyFn[P, T]) -> AnyFn[P, T]:
        if not iscoroutinefunction(fn):
            raise NotImplementedError(f'{fn} is sync')
        
        module = fn.__module__.replace('.','/')
        cache_path_for_fn = cache_base_path + module + '/' + fn.__name__

        # Ensure the cache dir exists
        os.makedirs(cache_path_for_fn, exist_ok=True)

        @toolcache.cache('disk', cache_dir=cache_path_for_fn)
        @functools.wraps(fn)
        async def diskcache_wrap(*args, **kwargs) -> T:
            logger.debug(f"fetching {fn.__qualname__}({', '.join(str(a) for a in args)})")
            return await fn(*args, **kwargs)
        return diskcache_wrap
        
except ImportError:

    """User doesn't have toolcache, this decorator will just return the undecorated function."""

    def optional_async_diskcache(fn: AnyFn[P, T]) -> AnyFn[P, T]:
        return fn



import abc
from typing import Generic, Optional, Tuple, Type, TypeVar

import a_sync

import y.ENVIRONMENT_VARIABLES as ENVS
from y import contracts
from y.classes._abc import LiquidityPool
from y._decorators import stuck_coro_debugger
from y.datatypes import AddressOrContract, AnyAddressType, Block, UsdPrice


class BalancerPool(LiquidityPool):
    """A :class:`~LiquidityPool` specific to the Balancer protocol."""

_B = TypeVar("_B", bound=BalancerPool)

class BalancerABC(a_sync.ASyncGenericBase, Generic[_B]):
    def __repr__(self) -> str:
        return f"<{type(self).__name__} object at {hex(id(self))}>"
    @a_sync.a_sync(ram_cache_ttl=5*60)
    @stuck_coro_debugger
    async def is_pool(self, token_address: AnyAddressType) -> bool:
        return await contracts.has_methods(token_address, self._check_methods, sync=False)
    @stuck_coro_debugger
    async def get_pool_price(self, pool_address: AnyAddressType, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> UsdPrice:
        return await self._pool_type(pool_address, asynchronous=True).get_pool_price(block=block, skip_cache=skip_cache)
    @abc.abstractproperty
    def _pool_type(self) -> Type[_B]:
        ...
    @abc.abstractproperty
    def _check_methods(self) -> Tuple[str]:
        ...
    @abc.abstractmethod
    async def get_token_price(self, token_address: AddressOrContract, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> Optional[UsdPrice]:
        ...



import logging
import time
from functools import lru_cache, wraps
from typing import Callable, Iterable, TypeVar
from typing_extensions import ParamSpec

import a_sync
from a_sync.a_sync import ASyncFunction
from brownie import chain
from pony.orm import (CommitException, OperationalError, TransactionError,
                      UnexpectedError, commit, db_session)

_T = TypeVar('_T')
_P = ParamSpec('_P')

logger = logging.getLogger(__name__)


ydb_read_threads = a_sync.PruningThreadPoolExecutor(16)
ydb_write_threads = a_sync.PruningThreadPoolExecutor(16)

def retry_locked(callable: Callable[_P, _T]) -> Callable[_P, _T]:
    @wraps(callable)
    def retry_locked_wrap(*args: _P.args, **kwargs: _P.kwargs) -> _T:
        sleep = 0.05
        while True:
            try:
                retval = callable(*args, **kwargs)
                commit()
                return retval
            except (CommitException, OperationalError, UnexpectedError) as e:
                log = logger.warning if sleep > 1 else logger.debug
                log("%s.%s got exc %s", callable.__module__, callable.__name__, e)
                if "database is locked" not in str(e):
                    raise
                time.sleep(sleep)
                sleep *= 1.5
            except TransactionError as e:
                logger.debug("%s.%s got exc %s", callable.__module__, callable.__name__, e)
                if "An attempt to mix objects belonging to different transactions" not in str(e):
                    raise
    return retry_locked_wrap

a_sync_read_db_session: Callable[[Callable[_P, _T]], ASyncFunction[_P, _T]] = lambda fn: a_sync.a_sync(default='async', executor=ydb_read_threads)(
    retry_locked(
        db_session(
            retry_locked(
                fn
            )
        )
    )
)

a_sync_write_db_session: Callable[[Callable[_P, _T]], ASyncFunction[_P, _T]] = lambda fn: a_sync.a_sync(default='async', executor=ydb_write_threads)(
    retry_locked(
        db_session(
            retry_locked(
                fn
            )
        )
    )
)

a_sync_write_db_session_cached: Callable[[Callable[_P, _T]], ASyncFunction[_P, _T]] = lambda fn: a_sync.a_sync(default='async', executor=ydb_write_threads, ram_cache_maxsize=None)(
    retry_locked(
        lru_cache(maxsize=None)(
            db_session(
                retry_locked(
                    fn
                )
            )
        )
    )
)


_result_count_logger = logging.getLogger(f"{__name__}.result_count")

def log_result_count(name: str, arg_names: Iterable[str] = []) -> Callable[[Callable[_P, _T]], Callable[_P, _T]]:
    def result_count_deco(fn: Callable[_P, _T]) -> Callable[_P, _T]:
        @wraps(fn)
        def result_count_wrap(*args: _P.args, **kwargs: _P.kwargs) -> _T:
            results = fn(*args, **kwargs)
            if _result_count_logger.isEnabledFor(logging.DEBUG):
                arg_values = ' '.join(f'{k} {v}' for k, v in [('chain', chain.id), *zip(arg_names, args)])
                _result_count_logger.debug("loaded %s %s for %s", len(results), name, arg_values)
            return results
        return result_count_wrap
    return result_count_deco



import asyncio
import logging
from typing import List, Optional, Set, Tuple

import a_sync
import dank_mids
import eth_retry
from a_sync.a_sync.property import HiddenMethodDescriptor
from brownie import chain
from multicall.call import Call
from typing_extensions import Self

from y._decorators import stuck_coro_debugger
from y.contracts import Contract, contract_creation_block_async
from y.datatypes import Address, AnyAddressType, Block
from y.interfaces.uniswap.velov2 import VELO_V2_FACTORY_ABI
from y.networks import Network
from y.prices.dex.solidly import SolidlyRouterBase
from y.prices.dex.uniswap.v2 import Path, UniswapV2Pool
from y.utils import gather_methods
from y.utils.cache import a_sync_ttl_cache
from y.utils.raw_calls import raw_call

_INIT_METHODS = 'token0()(address)', 'token1()(address)', 'stable()(bool)'

logger = logging.getLogger(__name__)

class NoReservesError(Exception):
    pass

default_factory = {
    Network.Optimism: "0xF1046053aa5682b4F9a81b5481394DA16BE5FF5a",
    Network.Base: "0x420DD381b31aEf6683db6B902084cB0FFECe40Da",
}

class VelodromePool(UniswapV2Pool):
    __slots__ = "is_stable", 
    def __init__(
        self, 
        address: AnyAddressType, 
        token0: Optional[AnyAddressType] = None, 
        token1: Optional[AnyAddressType] = None,
        stable: Optional[bool] = None,
        deploy_block: Optional[int] = None,
        asynchronous: bool = False,
    ):
        super().__init__(address, token0=token0, token1=token1, deploy_block=deploy_block, asynchronous=asynchronous)
        self.is_stable = stable

class VelodromeRouterV2(SolidlyRouterBase):
    _supports_uniswap_helper = False
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.default_factory = default_factory[chain.id]
    
    @stuck_coro_debugger
    @a_sync_ttl_cache
    async def pool_for(self, input_token: Address, output_token: Address, stable: bool) -> Optional[Address]:
        pool_address = str(await self.contract.poolFor.coroutine(input_token, output_token, stable, self.default_factory))
        if await is_contract(pool_address):
            return pool_address
    
    @stuck_coro_debugger
    @a_sync_ttl_cache
    @eth_retry.auto_retry
    async def get_pool(self, input_token: Address, output_token: Address, stable: bool, block: Block) -> Optional[UniswapV2Pool]:
        if pool_address := await self.pool_for(input_token, output_token, stable, sync=False):
            if await contract_creation_block_async(pool_address) <= block:
                return UniswapV2Pool(pool_address, asynchronous=self.asynchronous)
    
    @a_sync.aka.cached_property
    @stuck_coro_debugger
    async def pools(self) -> Set[VelodromePool]:
        logger.info('Fetching pools for %s on %s. If this is your first time using ypricemagic, this can take a while. Please wait patiently...', self.label, Network.printable())
        factory = await Contract.coroutine(self.factory)
        if 'PoolCreated' not in factory.topics:
            # the etherscan proxy detection is borked here, need this to decode properly
            factory = Contract.from_abi("PoolFactory", self.factory, VELO_V2_FACTORY_ABI)

        pools = {
            VelodromePool(
                address=event['pool'],
                token0=event['token0'], 
                token1=event['token1'], 
                stable=event['stable'], 
                deploy_block=event.block_number, 
                asynchronous=self.asynchronous,
            )
            async for event in factory.events.PoolCreated.events(to_block=await dank_mids.eth.block_number)
        }
        
        all_pools_len = await raw_call(self.factory, 'allPoolsLength()', output='int', sync=False)

        if len(pools) > all_pools_len:
            raise ValueError('wtf', len(pools), all_pools_len)
        
        if len(pools) < all_pools_len:
            logger.debug("Oh no! Looks like your node can't look back that far. Checking for the missing %s pools...", all_pools_len - len(pools))
            pools_your_node_couldnt_get = a_sync.map(self._init_pool_from_poolid, range(all_pools_len - len(pools)), name=f"load {self} poolId")
            # we want the map populated with tasks for this logger
            await pools_your_node_couldnt_get._init_loader
            logger.debug('pools: %s', pools_your_node_couldnt_get)
            pools.update(await pools_your_node_couldnt_get.values(pop=True))

        tokens = set()
        for pool in pools:
            tokens.update(await asyncio.gather(pool.__token0__, pool.__token1__))
        logger.info('Loaded %s pools supporting %s tokens on %s', len(pools), len(tokens), self.label)
        return pools
    __pools__: HiddenMethodDescriptor[Self, Set[VelodromePool]]
    
    @stuck_coro_debugger
    async def get_routes_from_path(self, path: Path, block: Block) -> List[Tuple[Address, Address, bool]]:
        routes = []
        for i in range(len(path) - 1):
            input_token, output_token = path[i], path[i+1]
            # Try for a stable pool first and use that if available
            stable_pool: Optional[VelodromePool]
            unstable_pool: Optional[VelodromePool]
            stable_pool, unstable_pool = await asyncio.gather(
                self.get_pool(input_token, output_token, True, block, sync=False),
                self.get_pool(input_token, output_token, False, block, sync=False),
            )
            
            if stable_pool and unstable_pool:
                # We have to find out which of these pools is deepest
                stable_reserves, unstable_reserves = await asyncio.gather(
                    stable_pool.reserves(block=block, sync=False),
                    unstable_pool.reserves(block=block, sync=False),
                )
                if stable_reserves and unstable_reserves:
                    stable_reserves = tuple(stable_reserves)
                    unstable_reserves = tuple(unstable_reserves)

                    # NOTE: using `__token0__` and `__token1__` is faster than `__tokens__` since they're already cached and return instantly
                    #       it also creates 2 fewer tasks and 1 fewer future than `__tokens__` since there is no use of `asyncio.gather`.
                    if await stable_pool.__token0__ == await unstable_pool.__token0__ and await stable_pool.__token1__ == await unstable_pool.__token1__:
                        stable_reserve = stable_reserves[0]
                        unstable_reserve = unstable_reserves[0]
                    else:  # Order of tokens is flip flopped in the pools
                        stable_reserve = stable_reserves[0]
                        unstable_reserve = unstable_reserves[1]
                    if stable_reserve >= unstable_reserve:
                        is_stable = True
                    elif stable_reserve < unstable_reserve:
                        is_stable = False
                    routes.append([input_token, output_token, is_stable, self.factory])
                elif stable_reserves:
                    routes.append([input_token, output_token, True, self.factory])
                elif unstable_reserves:
                    routes.append([input_token, output_token, False, self.factory])
                else:
                    raise NoReservesError(f"No route available for path {path}")
            elif stable_pool:
                routes.append([input_token, output_token, True, self.factory])
            elif unstable_pool:
                routes.append([input_token, output_token, False, self.factory])
            else:
                raise ValueError("Not sure why this function is even running if no pool is found")
            
        return routes

    @stuck_coro_debugger
    async def _init_pool_from_poolid(self, poolid: int) -> VelodromePool:
        logger.debug("initing poolid %s", poolid)
        pool = await Call(self.factory, ['allPools(uint256)(address)']).coroutine(poolid)
        if pool is None:  # TODO: debug why this happens sometimes and why this if clause works to get back on track
            factory = await Contract.coroutine(self.factory)
            pool = await factory.allPools.coroutine(poolid)
        token0, token1, stable = await gather_methods(pool, _INIT_METHODS)
        return VelodromePool(
            address=pool,
            token0=token0,
            token1=token1, 
            stable=stable, 
            asynchronous=self.asynchronous,
        )


async def is_contract(pool_address: Address) -> bool:
    if pool_address in __pools:
        return True
    if result := await dank_mids.eth.get_code(pool_address) not in ['0x',b'']:
        __pools.append(pool_address)
    return result

__pools = []


import asyncio
from typing import Optional

import a_sync
from a_sync.a_sync import HiddenMethodDescriptor
from brownie import chain
from brownie.exceptions import VirtualMachineError
from typing_extensions import Self

from y import Contract
from y.classes.common import ERC20
from y.datatypes import Address, AddressOrContract, Block
from y.exceptions import UnsupportedNetwork
from y.networks import Network

addresses = {
    # https://docs.fantom.foundation/tutorials/band-protocol-standard-dataset
    Network.Fantom: '0x56E2898E0ceFF0D1222827759B56B28Ad812f92F'
}

supported_assets = {
    # https://docs.fantom.foundation/tutorials/band-protocol-standard-dataset#supported-tokens
    Network.Fantom: [
        "0xaf319E5789945197e365E7f7fbFc56B130523B33", # FRAX
        "0x04068DA6C83AFCFA0e13ba15A6696662335D5B75", # USDC
        "0x0E1694483eBB3b74d3054E383840C6cf011e518e", # sUSD
        "0x6a07A792ab2965C72a5B8088d3a069A7aC3a993B", # AAVE
        "0x46E7628E8b4350b2716ab470eE0bA1fa9e76c6C5", # BAND
        "0xD67de0e0a0Fd7b15dC8348Bb9BE742F3c5850454", # BNB
        "0x321162Cd933E2Be498Cd2267a90534A804051b11", # BTC
        "0xB01E8419d842beebf1b70A7b5f7142abbaf7159D", # COVER
        "0x657A1861c15A3deD9AF0B6799a195a249ebdCbc6", # CREAM
        "0x1E4F97b9f9F913c46F1632781732927B9019C68b", # CRV
        "0x74b23882a30290451A17c44f4F05243b6b58C76d", # ETH
        "0x44B26E839eB3572c5E959F994804A5De66600349", # HEGIC
        "0x2A5062D22adCFaAfbd5C541d4dA82E4B450d4212", # KP3R
        "0xb3654dc3D10Ea7645f8319668E8F54d2574FBdC8", # LINK
        "0x924828a9Fb17d47D0eb64b57271D10706699Ff11", # SFI
        "0x56ee926bD8c72B2d5fa1aF4d9E4Cbb515a1E3Adc", # SNX
        "0xae75A438b2E0cB8Bb01Ec1E1e376De11D44477CC", # SUSHI
        "0x29b0Da86e484E1C0029B56e817912d778aC0EC69", # YFI
    ]
}

class Band(a_sync.ASyncGenericSingleton):
    def __init__(self, asynchronous: bool = False) -> None:
        if chain.id not in addresses:
            raise UnsupportedNetwork('band is not supported on this network')
        self.asynchronous = asynchronous
        super().__init__()

    def __contains__(self, asset: AddressOrContract) -> bool:
        return chain.id in addresses and asset in supported_assets[chain.id]
    
    @a_sync.aka.property
    async def oracle(self) -> Contract:
        return await Contract.coroutine(addresses[chain.id])
    __oracle__: HiddenMethodDescriptor[Self, Contract]

    async def get_price(self, asset: Address, block: Optional[Block] = None) -> Optional[float]:
        oracle, asset_symbol = await asyncio.gather(
            self.__oracle__,
            ERC20(asset, asynchronous=True).symbol,
        )
        try:
            reference_data = await oracle.getReferenceData.coroutine(asset_symbol, 'USDC', block_identifier=block)
            return reference_data[0] / 1e18
        except ValueError:
            return None
        except VirtualMachineError:
            return None


try:
    band = Band(asynchronous=True)
except UnsupportedNetwork:
    band = set()




import abc
import asyncio
import logging
from typing import (Any, AsyncIterator, Awaitable, Callable, Generic, List, NoReturn,
                    Optional, Type, TypeVar, Union)

import a_sync
import dank_mids
import eth_retry
from a_sync.executor import _AsyncExecutorMixin
from async_property import async_property
from brownie import ZERO_ADDRESS
from hexbytes import HexBytes
from pony.orm import (OptimisticCheckError, TransactionIntegrityError,
                      db_session)
from tqdm.asyncio import tqdm_asyncio
from web3.datastructures import AttributeDict
from web3.middleware.filter import block_ranges

from y import convert
from y._db.decorators import retry_locked
from y._db.exceptions import CacheNotPopulatedError
from y._decorators import stuck_coro_debugger
from y.utils.middleware import BATCH_SIZE

T = TypeVar('T')
S = TypeVar('S')
M = TypeVar('M')

logger = logging.getLogger(__name__)
default_filter_threads = a_sync.PruningThreadPoolExecutor(4)

def enc_hook(obj: Any) -> bytes:
    if isinstance(obj, AttributeDict):
        return dict(obj)
    elif isinstance(obj, HexBytes):
        return obj.hex()
    raise NotImplementedError(obj)

def dec_hook(typ: Type[T], obj: bytes) -> T:
    if typ == HexBytes:
        return HexBytes(obj)
    raise ValueError(f"{typ} is not a valid type for decoding")

class DiskCache(Generic[S, M], metaclass=abc.ABCMeta):
    __slots__ = []
    @abc.abstractmethod
    def _set_metadata(self, from_block: int, done_thru: int) -> None:
        """Updates the cache metadata to indicate the cache is populated from block `from_block` to block `to_block`."""
    @abc.abstractmethod
    def _is_cached_thru(self, from_block: int) -> int:
        """Returns max cached block for this cache or 0 if not cached."""
    @abc.abstractmethod
    def _select(self, from_block: int, to_block: int) -> List[S]:
        """Selects all cached objects from block `from_block` to block `to_block`"""
    @retry_locked
    def set_metadata(self, from_block: int, done_thru: int) -> None:
        """Updates the cache metadata to indicate the cache is populated from block `from_block` to block `to_block`."""
        try:
            with db_session:
                return self._set_metadata(from_block, done_thru)
        except TransactionIntegrityError as e:
            logger.debug("%s got exc %s when setting cache metadata", self, e)
            return self.set_metadata(from_block, done_thru)
        except OptimisticCheckError as e:
            # Don't need to update in this case
            logger.debug("%s got exc %s when setting cache metadata", self, e)
        
    @db_session
    @retry_locked
    def select(self, from_block: int, to_block: int) -> List[S]:
        """Selects all cached objects from block `from_block` to block `to_block`"""
        return self._select(from_block, to_block)
    @db_session
    @retry_locked
    def is_cached_thru(self, from_block: int) -> int:
        """Returns max cached block for this cache or 0 if not cached."""
        return self._is_cached_thru(from_block)
    def check_and_select(self, from_block: int, to_block: int) -> List[S]:
        """
        Selects all cached objects from block `from_block` to block `to_block` if the cache is fully populated.
        Raises `CacheNotPopulatedError` if it is not.
        """
        if self.is_cached_thru(from_block) >= to_block:
            return self.select(from_block, to_block)
        else:
            raise CacheNotPopulatedError(self, from_block, to_block)

C = TypeVar('C', bound=DiskCache)

class _DiskCachedMixin(a_sync.ASyncIterable[T], Generic[T, C], metaclass=abc.ABCMeta):
    __slots__ = 'is_reusable', '_cache', '_executor', '_objects', '_pruned'
    def __init__(
        self, 
        executor: Optional[_AsyncExecutorMixin] = None,
        is_reusable: bool = True,
    ):
        self.is_reusable = is_reusable
        self._cache = None
        self._executor = executor
        self._objects: List[T] = []
        self._pruned = 0
    @abc.abstractproperty
    def cache(self) -> C:
        ...
    @property
    def executor(self) -> _AsyncExecutorMixin:
        if self._executor is None:
            self._executor = default_filter_threads
        return self._executor

    @abc.abstractproperty
    def insert_to_db(self) -> Callable[[T], None]:
        ...
    
    #abc.abstractproperty
    def bulk_insert(self) -> Callable[[List[T]], Awaitable[None]]:
        ...
        
    async def _extend(self, objs) -> None:
        """Override this to pre-process objects before storing."""
        return self._objects.extend(objs)
    def _remove(self, obj: T) -> None:
        self._objects.remove(obj)
        self._pruned += 1
    async def _load_cache(self, from_block: int) -> int:
        """
        Loads cached logs from disk.
        Returns max block of logs loaded from cache.
        """
        logger.debug('checking to see if %s is cached in local db', self)
        if cached_thru := await self.executor.run(self.cache.is_cached_thru, from_block):
            logger.debug('%s is cached thru block %s, loading from db', self, cached_thru)
            await self._extend(await self.executor.run(self.cache.select, from_block, cached_thru))
            logger.debug('%s loaded %s objects thru block %s from disk', self, len(self._objects), cached_thru)
            return cached_thru
        return None

_E = TypeVar("_E", bound=_AsyncExecutorMixin)
    
class Filter(_DiskCachedMixin[T, C]):
    # defaults are stored as class vars to keep instance dicts smaller
    _chunk_size = BATCH_SIZE
    _chunks_per_batch = None
    _exc = None
    _db_task = None
    _sleep_fut = None
    _sleep_time = 60
    _task = None
    _semaphore = None
    _verbose = False
    __slots__ = 'from_block', 'to_block', '_interval', '_lock', '__dict__', '__weakref__'
    def __init__(
        self, 
        from_block: int,
        *, 
        chunk_size: int = BATCH_SIZE, 
        chunks_per_batch: Optional[int] = None,
        sleep_time: int = 60,
        semaphore: Optional[dank_mids.BlockSemaphore] = None,
        executor: Optional[_AsyncExecutorMixin] = None,
        is_reusable: bool = True,
        verbose: bool = False,
    ):
        self.from_block = from_block
        if chunk_size != self._chunk_size:
            self._chunk_size = chunk_size
        if chunks_per_batch != self._chunks_per_batch:
            self._chunks_per_batch = chunks_per_batch
        self._lock = a_sync.CounterLock()
        if semaphore != self._semaphore:
            self._semaphore = semaphore
        if sleep_time != self._sleep_time:
            self._sleep_time = sleep_time
        if verbose != self._verbose:
            self._verbose = verbose
        super().__init__(executor=executor, is_reusable=is_reusable)

    def __aiter__(self) -> AsyncIterator[T]:
        return self._objects_thru(block=None).__aiter__()
        
    def __del__(self) -> None:
        if self._task and not self._task.done():
            self._task.cancel()
        
    @abc.abstractmethod
    async def _fetch_range(self, from_block: int, to_block: int) -> List[T]:
        ...
        
    @property
    def semaphore(self) -> dank_mids.BlockSemaphore:
        if self._semaphore is None:
            self._semaphore = dank_mids.BlockSemaphore(self._chunks_per_batch)
        return self._semaphore

    @property
    def is_asleep(self) -> bool:
        return not self._sleep_fut.done() if self._sleep_fut else False
    
    def _get_block_for_obj(self, obj: T) -> int:
        """Override this as needed for different object types"""
        return obj['blockNumber']
    
    @a_sync.ASyncIterator.wrap
    async def _objects_thru(self, block: Optional[int]) -> AsyncIterator[T]:
        self._ensure_task()
        yielded = 0
        done_thru = 0
        while True:
            if block is None or done_thru < block:
                if self.is_asleep:
                    self._wakeup()
                await self._lock.wait_for(done_thru + 1)
            if self._exc:
                # create a new duplicate exc instead of building a massive traceback on the original
                try:
                    raise type(self._exc)(*self._exc.args).with_traceback(self._tb)
                except TypeError:
                    raise self._exc.with_traceback(self._tb) from None
            if to_yield := self._objects[yielded-self._pruned:]:
                for obj in to_yield:
                    if block and self._get_block_for_obj(obj) > block:
                        return
                    if not self.is_reusable:
                        self._remove(obj)
                    yield obj
                    yielded += 1
            elif block and done_thru >= block:
                return
            done_thru = self._lock.value
            logger.debug('%s lock value %s to_block %s', self, done_thru, block)
            if block is None:
                await asyncio.sleep(self._sleep_time)

    @async_property  
    async def _sleep(self) -> None:
        if self._sleep_fut is None or self._sleep_fut.done():
            self._sleep_fut = asyncio.get_event_loop().create_future()
        await self._sleep_fut
    
    def _wakeup(self) -> None:
        self._sleep_fut.set_result(None)
    
    async def __fetch(self) -> NoReturn:
        from y.constants import BIG_VALUE
        try:
            await self._fetch()
        except Exception as e:
            import traceback
            logger.exception(e)
            self._exc = e
            self._tb = e.__traceback__
            # no need to hold vars in memory
            traceback.clear_frames(self._tb)
            self._lock.set(BIG_VALUE)
            raise
    
    async def _fetch(self) -> NoReturn:
        """Override this if you want"""
        await self._loop(self.from_block)
    
    @stuck_coro_debugger
    async def _fetch_range_wrapped(self, i: int, range_start: int, range_end: int) -> List[T]:
        async with self.semaphore[range_end]:
            logger.debug("fetching %s block %s to %s", self, range_start, range_end)
            return i, range_end, await self._fetch_range(range_start, range_end)

    async def _loop(self, from_block: int) -> NoReturn:
        logger.debug('starting work loop for %s', self)
        if cached_thru := await self._load_cache(from_block):
            self._lock.set(cached_thru)
        while True:
            await self._load_new_objects(start_from_block=cached_thru or from_block)
            await self._sleep
    
    @eth_retry.auto_retry
    @stuck_coro_debugger
    async def _load_new_objects(self, to_block: Optional[int] = None, start_from_block: Optional[int] = None) -> None:
        logger.debug('loading new objects for %s', self)
        start = v + 1 if (v := self._lock.value) else start_from_block or self.from_block
        if to_block:
            end = to_block
            if start > end:
                raise ValueError(f"start {start} is bigger than end {end}, can't do that")
        else:
            while start > (end := await dank_mids.eth.block_number):
                logger.debug('%s start %s is greater than end %s, sleeping...', self, start, end)
                await asyncio.sleep(1)
        await self._load_range(start, end)

    @stuck_coro_debugger
    async def _load_range(self, from_block: int, to_block: int) -> None:
        logger.debug('loading block range %s to %s', from_block, to_block)
        chunks_yielded = 0
        done = {}
        coros = [
            self._fetch_range_wrapped(i, start, end) 
            for i, (start, end) 
            in enumerate(block_ranges(from_block, to_block, self._chunk_size))
            if self._chunks_per_batch is None or i < self._chunks_per_batch
        ]
        async for i, end, objs in a_sync.as_completed(coros, aiter=True, tqdm=self._verbose):
            next_chunk_loaded = False
            done[i] = end, objs
            for i in range(chunks_yielded, len(coros)):
                if i not in done:
                    break
                end, objs = done.pop(i)
                self._insert_chunk(objs, from_block, end)
                await self._extend(objs)
                next_chunk_loaded = True
                chunks_yielded += 1
            if next_chunk_loaded:
                await self._set_lock(end)
                logger.debug("%s loaded thru block %s", self, end)
    
    @stuck_coro_debugger
    async def _set_lock(self, block: int) -> None:
        """Override this if you want to, for things like awaiting for tasks to complete as I do in the curve module"""
        self._lock.set(block)
    
    def _insert_chunk(self, objs: List[T], from_block: int, done_thru: int) -> None:
        if (prev_task := self._db_task) and prev_task.done() and (e := prev_task.exception()):
            raise e
        depth = prev_task._depth + 1 if prev_task else 0
        logger.debug("%s queuing next db insert chunk %s thru block %s", self, depth, done_thru)
        task = asyncio.create_task(
            coro = self.__insert_chunk(objs, from_block, done_thru, prev_task, depth),
            name = f"_insert_chunk from {from_block} to {done_thru}",
        )
        task._depth = depth
        task._prev_task = prev_task
        self._db_task = task

    def _ensure_task(self) -> None:
        if self._task is None:
            logger.debug('creating task for %s', self)
            self._task = asyncio.create_task(coro=self.__fetch(), name=f"{self}.__fetch")
            # NOTE: The task does not return and will be cancelled when this object is 
            # garbage collected so there is no need to log the "destroy pending task" message.
            self._task._log_destroy_pending = False
        if self._task.done() and (e := self._task.exception()):
            raise e.with_traceback(e.__traceback__)
        
    @stuck_coro_debugger
    async def __insert_chunk(self, objs: List[T], from_block: int, done_thru: int, prev_chunk_task: Optional[asyncio.Task], depth: int) -> None:
        if prev_chunk_task:
            await prev_chunk_task
        if objs:
            await self.bulk_insert(objs, executor=self.executor)
        await self.executor.run(self.cache.set_metadata, from_block, done_thru)
        logger.debug("%s chunk %s thru block %s is now in db", self, depth, done_thru)
    
def _clean_addresses(addresses) -> Union[str, List[str]]:
    if addresses == ZERO_ADDRESS:
        raise ValueError("Cannot make a LogFilter for the zero address")
    if not addresses:
        return addresses
    if isinstance(addresses, str):
        return convert.to_address(addresses)
    elif hasattr(addresses, '__iter__'):
        if ZERO_ADDRESS in addresses:
            raise ValueError("Cannot make a LogFilter for the zero address")
        return [convert.to_address(address) for address in addresses]
    return convert.to_address(addresses)


import asyncio
import functools
import logging
from typing import (Callable, Dict, Iterable, List, Literal, 
                    Optional, Tuple, TypeVar, overload)

import a_sync
import dank_mids
from brownie import ZERO_ADDRESS
from brownie.exceptions import ContractNotFound
from typing_extensions import ParamSpec

from y import ENVIRONMENT_VARIABLES as ENVS
from y import constants, convert
from y._decorators import stuck_coro_debugger
from y.classes import ERC20
from y.datatypes import AnyAddressType, Block, Pool, UsdPrice
from y.exceptions import NonStandardERC20, PriceError, yPriceMagicError
from y.prices import band, chainlink, convex, one_to_one, pendle, popsicle, rkp3r, solidex, utils, yearn
from y.prices.dex import *
from y.prices.dex.uniswap import UniswapV2Pool, uniswap_multiplexer
from y.prices.eth_derivs import *
from y.prices.gearbox import gearbox
from y.prices.lending import *
from y.prices.stable_swap import *
from y.prices.synthetix import synthetix
from y.prices.tokenized_fund import *
from y.utils.logging import get_price_logger

_P = ParamSpec("_P")
_T = TypeVar("_T")

cache_logger = logging.getLogger(f"{__name__}.cache")


@overload
async def get_price(
    token_address: AnyAddressType,
    block: Optional[Block] = None,
    *,
    fail_to_None: Literal[True],
    skip_cache: bool = ENVS.SKIP_CACHE,
    ignore_pools: Tuple[Pool, ...] = (),
    silent: bool = False,
) -> Optional[UsdPrice]:...

@overload
async def get_price(
    token_address: AnyAddressType,
    block: Optional[Block] = None,
    *,
    fail_to_None: bool = False,
    skip_cache: bool = ENVS.SKIP_CACHE,
    ignore_pools: Tuple[Pool, ...] = (),
    silent: bool = False,
) -> UsdPrice:...

@a_sync.a_sync(default='sync')
async def get_price(
    token_address: AnyAddressType,
    block: Optional[Block] = None,
    *,
    fail_to_None: bool = False,
    skip_cache: bool = ENVS.SKIP_CACHE,
    ignore_pools: Tuple[Pool, ...] = (),
    silent: bool = False,
) -> Optional[UsdPrice]:
    """
    Get the price of a token in USD.

    Args:
        token_address: The address of the token to price.
        block (optional): The block number at which to get the price. If None, uses the latest block.
        fail_to_None (optional): If True, return None instead of raising a :class:`~yPriceMagicError` on failure. Default False.
        skip_cache (optional): If True, bypass the cache and fetch the price directly. Defaults to :obj:`ENVS.SKIP_CACHE`.
        ignore_pools (optional): A tuple of pool addresses to ignore when fetching the price.
        silent (optional): If True, suppress error logging. Default False.

    Returns:
        The price of the token in USD, or None if the price couldn't be determined and fail_to_None is True.

    Raises:
        yPriceMagicError: If the price couldn't be determined and fail_to_None is False.
    
    Note:
        Don't pass an int like `123` into `token_address` please, that's just silly.
        - ypricemagic accepts ints to allow you to pass `y.get_price(0x0bc529c00C6401aEF6D220BE8C6Ea1667F6Ad93e)`
            so you can save yourself some keystrokes while testing in a console
        - (as opposed to `y.get_price("0x0bc529c00C6401aEF6D220BE8C6Ea1667F6Ad93e")`)
    """
    block = block or await dank_mids.eth.block_number
    token_address = convert.to_address(token_address)
    try:
        return await _get_price(token_address, block, fail_to_None=fail_to_None, ignore_pools=ignore_pools, skip_cache=skip_cache, silent=silent)
    except (ContractNotFound, NonStandardERC20, PriceError) as e:
        symbol = await ERC20(token_address, asynchronous=True).symbol
        if not fail_to_None:
            raise_from = None if isinstance(e, PriceError) else e
            raise yPriceMagicError(e, token_address, block, symbol) from raise_from



@overload
async def get_prices(
    token_addresses: Iterable[AnyAddressType],
    block: Optional[Block] = None,
    *,
    fail_to_None: Literal[True],
    skip_cache: bool = ENVS.SKIP_CACHE,
    silent: bool = False,
) -> List[Optional[UsdPrice]]:...

@overload
async def get_prices(
    token_addresses: Iterable[AnyAddressType],
    block: Optional[Block] = None,
    *,
    fail_to_None: bool = False,
    skip_cache: bool = ENVS.SKIP_CACHE,
    silent: bool = False,
) -> List[UsdPrice]:...

@a_sync.a_sync(default='sync')
async def get_prices(
    token_addresses: Iterable[AnyAddressType],
    block: Optional[Block] = None,
    *,
    fail_to_None: bool = False,
    skip_cache: bool = ENVS.SKIP_CACHE,
    silent: bool = False,
) -> List[Optional[UsdPrice]]:
    """
    Get prices for multiple tokens in USD.

    You should use this function over :func:`get_price` where possible, it is better optimized for parallel execution.

    Args:
        token_addresses: An iterable of token addresses to price.
        block (optional): The block number at which to get the prices. Defaults to the latest block.
        fail_to_None (optional): If True, return None for tokens whose price couldn't be determined. Default False.
        skip_cache (optional): If True, bypass the cache and fetch prices directly. Defaults to :obj:`ENVS.SKIP_CACHE`.
        silent (optional): If True, suppress progress bar and error logging. This kwarg is not currently implemented.

    Returns:
        A list of token prices in USD, in the same order as the input token_addresses.
    """
    return await map_prices(
        token_addresses, 
        block or await dank_mids.eth.block_number, 
        fail_to_None=fail_to_None, 
        skip_cache=skip_cache, 
        silent=silent,
    ).values(pop=True)



@overload
def map_prices(
    token_addresses: Iterable[AnyAddressType],
    block: Block,
    *,
    fail_to_None: Literal[True],
    skip_cache: bool = ENVS.SKIP_CACHE,
    silent: bool = False,
) -> a_sync.TaskMapping[AnyAddressType, Optional[UsdPrice]]:...

@overload
def map_prices(
    token_addresses: Iterable[AnyAddressType],
    block: Block,
    *,
    fail_to_None: bool = False,
    skip_cache: bool = ENVS.SKIP_CACHE,
    silent: bool = False,
) -> a_sync.TaskMapping[AnyAddressType, UsdPrice]:...

def map_prices(
    token_addresses: Iterable[AnyAddressType],
    block: Block,
    *,
    fail_to_None: bool = False,
    skip_cache: bool = ENVS.SKIP_CACHE,
    silent: bool = False,
) -> a_sync.TaskMapping[AnyAddressType, Optional[UsdPrice]]:
    """
    Map token addresses to their prices asynchronously.

    Args:
        token_addresses: An iterable of token addresses to price.
        block (optional): The block number at which to get the prices. Defaults to latest block.
        fail_to_None (optional): If True, map to None for tokens whose price couldn't be determined. Default False.
        skip_cache (optional): If True, bypass the cache and fetch prices directly. Defaults to :obj:`ENVS.SKIP_CACHE`.
        silent (optional): If True, suppress error logging. Default False.

    Returns:
        An :class:`a_sync.TaskMapping` object mapping token addresses to their prices.
    """
    return a_sync.map(
        get_price,
        token_addresses,
        block=block, 
        fail_to_None=fail_to_None, 
        skip_cache=skip_cache, 
        silent=silent, 
    )



def __cache(get_price: Callable[_P, _T]) -> Callable[_P, _T]:
    @functools.wraps(get_price)
    async def cache_wrap(
        token: AnyAddressType, 
        block: Block,
        *,
        fail_to_None: bool = False, 
        skip_cache: bool = ENVS.SKIP_CACHE,
        ignore_pools: Tuple[Pool, ...] = (),
        silent: bool = False
    ) -> Optional[UsdPrice]:
        from y._db.utils import price as db
        if not skip_cache and (price := await db.get_price(token, block)):
            cache_logger.debug('disk cache -> %s', price)
            return price
        price = await get_price(token, block=block, fail_to_None=fail_to_None, ignore_pools=ignore_pools, silent=silent)
        if price and not skip_cache:
            db.set_price(token, block, price)
        return price
    return cache_wrap

@stuck_coro_debugger
@a_sync.a_sync(default="async", cache_type='memory', ram_cache_ttl=ENVS.CACHE_TTL)
@__cache
async def _get_price(
    token: AnyAddressType, 
    block: Block,
    *,
    fail_to_None: bool = False, 
    skip_cache: bool = ENVS.SKIP_CACHE,
    ignore_pools: Tuple[Pool, ...] = (),
    silent: bool = False
    ) -> Optional[UsdPrice]:  # sourcery skip: remove-redundant-if
    """
    Internal function to get the price of a token.

    This function implements the core logic for fetching token prices.

    Args:
        token: The address of the token to price.
        block: The block number at which to get the price.
        fail_to_None: If True, return None instead of raising an exception on failure.
        skip_cache: If True, bypass the cache and fetch the price directly.
        ignore_pools: A tuple of pool addresses to ignore when fetching the price.
        silent: If True, suppress error logging.

    Returns:
        The price of the token in USD, or None if the price couldn't be determined and fail_to_None is True.
    """
    if token == ZERO_ADDRESS:
        _fail_appropriately(logger, symbol, fail_to_None, silent)
        return None
    
    try:
        # We do this to cache the symbol for later, otherwise some repr woudl break
        symbol = await ERC20(token, asynchronous=True).symbol
    except NonStandardERC20:
        symbol = None

    logger = get_price_logger(token, block, symbol=symbol, extra='magic', start_task=True)
    logger.debug('fetching price for %s', symbol)
    try:
        price = await _get_price_from_api(token, block, logger)
        if price is None:
            price = await _exit_early_for_known_tokens(token, block=block, ignore_pools=ignore_pools, skip_cache=skip_cache, logger=logger)
        if price is None:
            price = await _get_price_from_dexes(token, block, ignore_pools, skip_cache, logger)
        if price:
            await utils.sense_check(token, block, price)
        else:
            _fail_appropriately(logger, symbol, fail_to_None, silent)
        logger.debug("%s price: %s", symbol, price)
        if price:  # checks for the erroneous 0 value we see once in a while
            return price
    finally:
        logger.close()

@stuck_coro_debugger
async def _exit_early_for_known_tokens(
    token_address: str,
    block: Block,
    logger: logging.Logger,
    skip_cache: bool = ENVS.SKIP_CACHE,
    ignore_pools: Tuple[Pool, ...] = (),
    ) -> Optional[UsdPrice]:  # sourcery skip: low-code-quality
    """
    Attempt to get the price for known token types without having to fully load everything.

    This function checks if the token is of a known type (e.g., atoken, balancer pool, etc.)
    and attempts to get its price using type-specific methods.

    Args:
        token_address: The address of the token to price.
        block: The block number at which to get the price.
        logger: A logger instance for recording debug information.
        skip_cache: If True, bypass the cache and fetch the price directly.
        ignore_pools: A tuple of pool addresses to ignore when fetching the price.

    Returns:
        The price of the token if it can be determined early, or None otherwise.
    """
    bucket = await utils.check_bucket(token_address, sync=False)

    price = None

    if bucket == 'atoken':                  price = await aave.get_price(token_address, block=block, skip_cache=skip_cache, sync=False)
    elif bucket == 'balancer pool':         price = await balancer_multiplexer.get_price(token_address, block, skip_cache=skip_cache, sync=False)
    elif bucket == 'basketdao':             price = await basketdao.get_price(token_address, block, skip_cache=skip_cache, sync=False)

    elif bucket == 'belt lp':               price = await belt.get_price(token_address, block, sync=False)
    elif bucket == 'chainlink and band':    price = await chainlink.get_price(token_address, block, sync=False) or await band.get_price(token_address, block, sync=False)
    elif bucket == 'chainlink feed':        price = await chainlink.get_price(token_address, block, sync=False)

    elif bucket == 'compound':              price = await compound.get_price(token_address, block=block, skip_cache=skip_cache, sync=False)
    elif bucket == 'convex':                price = await convex.get_price(token_address, block, skip_cache=skip_cache, sync=False)
    elif bucket == 'creth':                 price = await creth.get_price_creth(token_address, block, skip_cache=skip_cache, sync=False)

    elif bucket == 'curve lp':              price = await curve.get_price(token_address, block, skip_cache=skip_cache, sync=False)
    elif bucket == 'ellipsis lp':           price = await ellipsis.get_price(token_address, block=block, skip_cache=skip_cache, sync=False)
    elif bucket == 'froyo':                 price = await froyo.get_price(token_address, block=block, sync=False)

    elif bucket == 'gearbox':               price = await gearbox.get_price(token_address, block=block, skip_cache=skip_cache, sync=False)
    elif bucket == 'gelato':                price = await gelato.get_price(token_address, block=block, skip_cache=skip_cache, sync=False)
    elif bucket == 'generic amm':           price = await generic_amm.get_price(token_address, block=block, skip_cache=skip_cache, sync=False)
    
    elif bucket == 'ib token':              price = await ib.get_price(token_address, block=block, skip_cache=skip_cache, sync=False)
    elif bucket == 'mooniswap lp':          price = await mooniswap.get_pool_price(token_address, block=block, skip_cache=skip_cache, sync=False)
    elif bucket == 'mstable feeder pool':   price = await mstablefeederpool.get_price(token_address,block=block, skip_cache=skip_cache, sync=False)
    
    elif bucket == 'one to one':            price = await one_to_one.get_price(token_address, block, skip_cache=skip_cache, sync=False)
    elif bucket == 'pendle lp':             price = await pendle.get_lp_price(token_address, block, skip_cache=skip_cache, sync=False)
    elif bucket == 'piedao lp':             price = await piedao.get_price(token_address, block=block, skip_cache=skip_cache, sync=False)

    elif bucket == 'popsicle':              price = await popsicle.get_price(token_address, block=block, skip_cache=skip_cache, sync=False)
    elif bucket == 'reserve':               price = await reserve.get_price(token_address, block=block, skip_cache=skip_cache, sync=False)
    elif bucket == 'rkp3r':                 price = await rkp3r.get_price(token_address, block, skip_cache=skip_cache, sync=False)

    elif bucket == 'saddle':                price = await saddle.get_price(token_address, block, skip_cache=skip_cache, sync=False)
    elif bucket == 'solidex':               price = await solidex.get_price(token_address, block, skip_cache=skip_cache, sync=False)
    elif bucket == 'stable usd':            price = 1

    elif bucket == 'synthetix':             price = await synthetix.get_price(token_address, block, sync=False)
    elif bucket == 'token set':             price = await tokensets.get_price(token_address, block=block, skip_cache=skip_cache, sync=False)
    elif bucket == 'uni or uni-like lp':    price = await UniswapV2Pool(token_address).get_price(block=block, skip_cache=skip_cache, sync=False)

    elif bucket == 'wrapped gas coin':      price = await get_price(constants.WRAPPED_GAS_COIN, block, skip_cache=skip_cache, sync=False)
    elif bucket == 'wrapped atoken v2':     price = await aave.get_price_wrapped_v2(token_address, block, skip_cache=skip_cache, sync=False)
    elif bucket == 'wrapped atoken v3':     price = await aave.get_price_wrapped_v3(token_address, block, skip_cache=skip_cache, sync=False)

    elif bucket == 'wsteth':                price = await wsteth.wsteth.get_price(block, skip_cache=skip_cache, sync=False)
    elif bucket == 'yearn or yearn-like':   price = await yearn.get_price(token_address, block, skip_cache=skip_cache, ignore_pools=ignore_pools, sync=False)

    logger.debug("%s -> %s", bucket, price)
    return price

async def _get_price_from_api(token: AnyAddressType, block: Block, logger: logging.Logger):
    """
    Attempt to get the price from the ypricemagic API.

    Args:
        token: The address of the token to price.
        block: The block number at which to get the price.
        logger: A logger instance for recording debug information.

    Returns:
        The price of the token if it can be fetched from the ypricemagic API, or None otherwise.
    """
    if utils.ypriceapi.should_use and token not in utils.ypriceapi.skip_tokens:
        price = await utils.ypriceapi.get_price(token, block)
        logger.debug("ypriceapi -> %s", price)
        return price

async def _get_price_from_dexes(token: AnyAddressType, block: Block, ignore_pools, skip_cache: bool, logger: logging.Logger):
    """
    Attempt to get the price from decentralized exchanges.

    This function tries to fetch the price from various DEXes like Uniswap, Curve, and Balancer.

    Args:
        token: The address of the token to price.
        block : The block number at which to get the price.
        ignore_pools: A tuple of pool addresses to ignore when fetching the price.
        skip_cache: If True, bypass the cache and fetch the price directly.
        logger: A logger instance for recording debug information.

    Returns:
        The price of the token if it can be determined from DEXes, or None otherwise.
    """
    # TODO We need better logic to determine whether to use uniswap, curve, balancer. For now this works for all known cases.
    dexes = [uniswap_multiplexer]
    if curve:
        dexes.append(curve)
        
    # TODO: make a DexABC, include balancer and future dexes
    # TODO:  this would be so cool if a_sync.map could proxy abstractmethods correctly
    # dexes_by_depth = dict(
    #     await DexABC.check_liquidity.map(dexes, token=token, block=block, ignore_pools=ignore_pools).items(pop=True).sort(lambda k, v: v)
    # )
    liquidity = await asyncio.gather(*[dex.check_liquidity(token, block, ignore_pools=ignore_pools, sync=False) for dex in dexes])
    depth_to_dex: Dict[int, object] = dict(zip(liquidity, dexes))
    dexes_by_depth: Dict[int, object] = {depth: depth_to_dex[depth] for depth in sorted(depth_to_dex, reverse=True) if depth}
    logger.debug('dexes by depth: %s', dexes_by_depth)
    for dex in dexes_by_depth.values():
        method = 'get_price_for_underlying' if hasattr(dex, 'get_price_for_underlying') else 'get_price'
        logger.debug("trying %s", dex)
        price = await getattr(dex, method)(token, block, ignore_pools=ignore_pools, skip_cache=skip_cache, sync=False)
        logger.debug("%s -> %s", dex, price)
        if price:
            return price
    logger.debug('no %s liquidity found on primary markets', token)

    # If price is 0, we can at least try to see if balancer gives us a price. If not, its probably a shitcoin.
    if price := await balancer_multiplexer.get_price(token, block=block, skip_cache=skip_cache, sync=False):
        logger.debug("balancer -> %s", price)
        return price
         
def _fail_appropriately(
    logger: logging.Logger,
    symbol: str, 
    fail_to_None: bool, 
    silent: bool,
    ) -> None:
    """
    Handle failure to get a price appropriately.

    This function decides how to handle a failure to get a price based on the input parameters.

    Args:
        logger: A logger instance for recording error information.
        symbol: The symbol of the token whose price couldn't be determined.
        fail_to_None: If True, the function will return silently. If False, it will raise a PriceError.
        silent: If True, suppress error logging.

    Raises:
        PriceError: If fail_to_None is False.
    """
    if not silent:
        logger.warning(f"failed to get price for {symbol}")

    if not fail_to_None:
        raise PriceError(logger, symbol)
